# Copyright 2022 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json
import logging
import os
import threading
from datetime import datetime, timedelta
from typing import NamedTuple, Optional

import dateutil.parser
from dateutil.tz import tzutc

from botocore import UNSIGNED
from botocore.compat import total_seconds
from botocore.config import Config
from botocore.exceptions import (
    ClientError,
    InvalidConfigError,
    TokenRetrievalError,
)
from botocore.utils import CachedProperty, JSONFileCache, SSOTokenLoader

logger = logging.getLogger(__name__)


def _utc_now():
    return datetime.now(tzutc())


def create_token_resolver(session):
    providers = [
        SSOTokenProvider(session),
    ]
    return TokenProviderChain(providers=providers)


def _serialize_utc_timestamp(obj):
    if isinstance(obj, datetime):
        return obj.strftime("%Y-%m-%dT%H:%M:%SZ")
    return obj


def _sso_json_dumps(obj):
    return json.dumps(obj, default=_serialize_utc_timestamp)


class FrozenAuthToken(NamedTuple):
    token: str
    expiration: Optional[datetime] = None


class DeferredRefreshableToken:
    # The time at which we'll attempt to refresh, but not block if someone else
    # is refreshing.
    _advisory_refresh_timeout = 15 * 60
    # The time at which all threads will block waiting for a refreshed token
    _mandatory_refresh_timeout = 10 * 60
    # Refresh at most once every minute to avoid blocking every request
    _attempt_timeout = 60

    def __init__(self, method, refresh_using, time_fetcher=_utc_now):
        self._time_fetcher = time_fetcher
        self._refresh_using = refresh_using
        self.method = method

        # The frozen token is protected by this lock
        self._refresh_lock = threading.Lock()
        self._frozen_token = None
        self._next_refresh = None

    def get_frozen_token(self):
        self._refresh()
        return self._frozen_token

    def _refresh(self):
        # If we don't need to refresh just return
        refresh_type = self._should_refresh()
        if not refresh_type:
            return None

        # Block for refresh if we're in the mandatory refresh window
        block_for_refresh = refresh_type == "mandatory"
        if self._refresh_lock.acquire(block_for_refresh):
            try:
                self._protected_refresh()
            finally:
                self._refresh_lock.release()

    def _protected_refresh(self):
        # This should only be called after acquiring the refresh lock
        # Another thread may have already refreshed, double check refresh
        refresh_type = self._should_refresh()
        if not refresh_type:
            return None

        try:
            now = self._time_fetcher()
            self._next_refresh = now + timedelta(seconds=self._attempt_timeout)
            self._frozen_token = self._refresh_using()
        except Exception:
            logger.warning(
                "Refreshing token failed during the %s refresh period.",
                refresh_type,
                exc_info=True,
            )
            if refresh_type == "mandatory":
                # This refresh was mandatory, error must be propagated back
                raise

        if self._is_expired():
            # Fresh credentials should never be expired
            raise TokenRetrievalError(
                provider=self.method,
                error_msg="Token has expired and refresh failed",
            )

    def _is_expired(self):
        if self._frozen_token is None:
            return False

        expiration = self._frozen_token.expiration
        remaining = total_seconds(expiration - self._time_fetcher())
        return remaining <= 0

    def _should_refresh(self):
        if self._frozen_token is None:
            # We don't have a token yet, mandatory refresh
            return "mandatory"

        expiration = self._frozen_token.expiration
        if expiration is None:
            # No expiration, so assume we don't need to refresh.
            return None

        now = self._time_fetcher()
        if now < self._next_refresh:
            return None

        remaining = total_seconds(expiration - now)

        if remaining < self._mandatory_refresh_timeout:
            return "mandatory"
        elif remaining < self._advisory_refresh_timeout:
            return "advisory"

        return None


class TokenProviderChain:
    def __init__(self, providers=None):
        if providers is None:
            providers = []
        self._providers = providers

    def load_token(self):
        for provider in self._providers:
            token = provider.load_token()
            if token is not None:
                return token
        return None


class SSOTokenProvider:
    METHOD = "sso"
    _REFRESH_WINDOW = 15 * 60
    _SSO_TOKEN_CACHE_DIR = os.path.expanduser(
        os.path.join("~", ".aws", "sso", "cache")
    )
    _SSO_CONFIG_VARS = [
        "sso_start_url",
        "sso_region",
    ]
    _GRANT_TYPE = "refresh_token"
    DEFAULT_CACHE_CLS = JSONFileCache

    def __init__(
        self, session, cache=None, time_fetcher=_utc_now, profile_name=None
    ):
        self._session = session
        if cache is None:
            cache = self.DEFAULT_CACHE_CLS(
                self._SSO_TOKEN_CACHE_DIR,
                dumps_func=_sso_json_dumps,
            )
        self._now = time_fetcher
        self._cache = cache
        self._token_loader = SSOTokenLoader(cache=self._cache)
        self._profile_name = (
            profile_name
            or self._session.get_config_variable("profile")
            or 'default'
        )

    def _load_sso_config(self):
        loaded_config = self._session.full_config
        profiles = loaded_config.get("profiles", {})
        sso_sessions = loaded_config.get("sso_sessions", {})
        profile_config = profiles.get(self._profile_name, {})

        if "sso_session" not in profile_config:
            return

        sso_session_name = profile_config["sso_session"]
        sso_config = sso_sessions.get(sso_session_name, None)

        if not sso_config:
            error_msg = (
                f'The profile "{self._profile_name}" is configured to use the SSO '
                f'token provider but the "{sso_session_name}" sso_session '
                f"configuration does not exist."
            )
            raise InvalidConfigError(error_msg=error_msg)

        missing_configs = []
        for var in self._SSO_CONFIG_VARS:
            if var not in sso_config:
                missing_configs.append(var)

        if missing_configs:
            error_msg = (
                f'The profile "{self._profile_name}" is configured to use the SSO '
                f"token provider but is missing the following configuration: "
                f"{missing_configs}."
            )
            raise InvalidConfigError(error_msg=error_msg)

        return {
            "session_name": sso_session_name,
            "sso_region": sso_config["sso_region"],
            "sso_start_url": sso_config["sso_start_url"],
        }

    @CachedProperty
    def _sso_config(self):
        return self._load_sso_config()

    @CachedProperty
    def _client(self):
        config = Config(
            region_name=self._sso_config["sso_region"],
            signature_version=UNSIGNED,
        )
        return self._session.create_client("sso-oidc", config=config)

    def _attempt_create_token(self, token):
        response = self._client.create_token(
            grantType=self._GRANT_TYPE,
            clientId=token["clientId"],
            clientSecret=token["clientSecret"],
            refreshToken=token["refreshToken"],
        )
        expires_in = timedelta(seconds=response["expiresIn"])
        new_token = {
            "startUrl": self._sso_config["sso_start_url"],
            "region": self._sso_config["sso_region"],
            "accessToken": response["accessToken"],
            "expiresAt": self._now() + expires_in,
            # Cache the registration alongside the token
            "clientId": token["clientId"],
            "clientSecret": token["clientSecret"],
            "registrationExpiresAt": token["registrationExpiresAt"],
        }
        if "refreshToken" in response:
            new_token["refreshToken"] = response["refreshToken"]
        logger.info("SSO Token refresh succeeded")
        return new_token

    def _refresh_access_token(self, token):
        keys = (
            "refreshToken",
            "clientId",
            "clientSecret",
            "registrationExpiresAt",
        )
        missing_keys = [k for k in keys if k not in token]
        if missing_keys:
            msg = f"Unable to refresh SSO token: missing keys: {missing_keys}"
            logger.info(msg)
            return None

        expiry = dateutil.parser.parse(token["registrationExpiresAt"])
        if total_seconds(expiry - self._now()) <= 0:
            logger.info(f"SSO token registration expired at {expiry}")
            return None

        try:
            return self._attempt_create_token(token)
        except ClientError:
            logger.warning("SSO token refresh attempt failed", exc_info=True)
            return None

    def _refresher(self):
        start_url = self._sso_config["sso_start_url"]
        session_name = self._sso_config["session_name"]
        logger.info(f"Loading cached SSO token for {session_name}")
        token_dict = self._token_loader(start_url, session_name=session_name)
        expiration = dateutil.parser.parse(token_dict["expiresAt"])
        logger.debug(f"Cached SSO token expires at {expiration}")

        remaining = total_seconds(expiration - self._now())
        if remaining < self._REFRESH_WINDOW:
            new_token_dict = self._refresh_access_token(token_dict)
            if new_token_dict is not None:
                token_dict = new_token_dict
                expiration = token_dict["expiresAt"]
                self._token_loader.save_token(
                    start_url, token_dict, session_name=session_name
                )

        return FrozenAuthToken(
            token_dict["accessToken"], expiration=expiration
        )

    def load_token(self):
        if self._sso_config is None:
            return None

        return DeferredRefreshableToken(
            self.METHOD, self._refresher, time_fetcher=self._now
        )



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import datetime
import itertools

import pytest
from dateutil.tz import tzutc

from botocore import model, parsers
from botocore.compat import MutableMapping, json
from tests import RawResponse, unittest


# HTTP responses will typically return a custom HTTP
# dict.  We want to ensure we're able to work with any
# kind of mutable mapping implementation.
class CustomHeaderDict(MutableMapping):
    def __init__(self, original_dict):
        self._d = original_dict

    def __getitem__(self, item):
        return self._d[item]

    def __setitem__(self, item, value):
        self._d[item] = value

    def __delitem__(self, item):
        del self._d[item]

    def __iter__(self):
        return iter(self._d)

    def __len__(self):
        return len(self._d)


# These tests contain botocore specific tests that either
# don't make sense in the protocol tests or haven't been added
# yet.
class TestResponseMetadataParsed(unittest.TestCase):
    def test_response_metadata_parsed_for_query_service(self):
        parser = parsers.QueryParser()
        response = (
            b'<OperationNameResponse>'
            b'  <OperationNameResult><Str>myname</Str></OperationNameResult>'
            b'  <ResponseMetadata>'
            b'    <RequestId>request-id</RequestId>'
            b'  </ResponseMetadata>'
            b'</OperationNameResponse>'
        )
        output_shape = model.StructureShape(
            'OutputShape',
            {
                'type': 'structure',
                'resultWrapper': 'OperationNameResult',
                'members': {
                    'Str': {
                        'shape': 'StringType',
                    },
                    'Num': {
                        'shape': 'IntegerType',
                    },
                },
            },
            model.ShapeResolver(
                {
                    'StringType': {
                        'type': 'string',
                    },
                    'IntegerType': {
                        'type': 'integer',
                    },
                }
            ),
        )
        parsed = parser.parse(
            {'body': response, 'headers': {}, 'status_code': 200}, output_shape
        )
        self.assertEqual(
            parsed,
            {
                'Str': 'myname',
                'ResponseMetadata': {
                    'RequestId': 'request-id',
                    'HTTPStatusCode': 200,
                    'HTTPHeaders': {},
                },
            },
        )

    def test_metadata_always_exists_for_query(self):
        # ResponseMetadata is used for more than just the request id. It
        # should always get populated, even if the request doesn't seem to
        # have an id.
        parser = parsers.QueryParser()
        response = (
            b'<OperationNameResponse>'
            b'  <OperationNameResult><Str>myname</Str></OperationNameResult>'
            b'</OperationNameResponse>'
        )
        output_shape = model.StructureShape(
            'OutputShape',
            {
                'type': 'structure',
                'resultWrapper': 'OperationNameResult',
                'members': {
                    'Str': {
                        'shape': 'StringType',
                    },
                    'Num': {
                        'shape': 'IntegerType',
                    },
                },
            },
            model.ShapeResolver(
                {
                    'StringType': {
                        'type': 'string',
                    },
                    'IntegerType': {
                        'type': 'integer',
                    },
                }
            ),
        )
        parsed = parser.parse(
            {'body': response, 'headers': {}, 'status_code': 200}, output_shape
        )
        expected = {
            'Str': 'myname',
            'ResponseMetadata': {'HTTPStatusCode': 200, 'HTTPHeaders': {}},
        }
        self.assertEqual(parsed, expected)

    def test_response_metadata_parsed_for_ec2(self):
        parser = parsers.EC2QueryParser()
        response = (
            b'<OperationNameResponse>'
            b'  <Str>myname</Str>'
            b'  <requestId>request-id</requestId>'
            b'</OperationNameResponse>'
        )
        output_shape = model.StructureShape(
            'OutputShape',
            {
                'type': 'structure',
                'members': {
                    'Str': {
                        'shape': 'StringType',
                    }
                },
            },
            model.ShapeResolver({'StringType': {'type': 'string'}}),
        )
        parsed = parser.parse(
            {'headers': {}, 'body': response, 'status_code': 200}, output_shape
        )
        # Note that the response metadata is normalized to match the query
        # protocol, even though this is not how it appears in the output.
        self.assertEqual(
            parsed,
            {
                'Str': 'myname',
                'ResponseMetadata': {
                    'RequestId': 'request-id',
                    'HTTPStatusCode': 200,
                    'HTTPHeaders': {},
                },
            },
        )

    def test_metadata_always_exists_for_ec2(self):
        # ResponseMetadata is used for more than just the request id. It
        # should always get populated, even if the request doesn't seem to
        # have an id.
        parser = parsers.EC2QueryParser()
        response = (
            b'<OperationNameResponse>'
            b'  <Str>myname</Str>'
            b'</OperationNameResponse>'
        )
        output_shape = model.StructureShape(
            'OutputShape',
            {
                'type': 'structure',
                'members': {
                    'Str': {
                        'shape': 'StringType',
                    }
                },
            },
            model.ShapeResolver({'StringType': {'type': 'string'}}),
        )
        parsed = parser.parse(
            {'headers': {}, 'body': response, 'status_code': 200}, output_shape
        )
        expected = {
            'Str': 'myname',
            'ResponseMetadata': {'HTTPStatusCode': 200, 'HTTPHeaders': {}},
        }
        self.assertEqual(parsed, expected)

    def test_response_metadata_on_json_request(self):
        parser = parsers.JSONParser()
        response = b'{"Str": "mystring"}'
        headers = {'x-amzn-requestid': 'request-id'}
        output_shape = model.StructureShape(
            'OutputShape',
            {
                'type': 'structure',
                'members': {
                    'Str': {
                        'shape': 'StringType',
                    }
                },
            },
            model.ShapeResolver({'StringType': {'type': 'string'}}),
        )
        parsed = parser.parse(
            {'body': response, 'headers': headers, 'status_code': 200},
            output_shape,
        )
        # Note that the response metadata is normalized to match the query
        # protocol, even though this is not how it appears in the output.
        self.assertEqual(
            parsed,
            {
                'Str': 'mystring',
                'ResponseMetadata': {
                    'RequestId': 'request-id',
                    'HTTPStatusCode': 200,
                    'HTTPHeaders': headers,
                },
            },
        )

    def test_response_no_initial_event_stream(self):
        parser = parsers.JSONParser()
        output_shape = model.StructureShape(
            'OutputShape',
            {
                'type': 'structure',
                'members': {'Payload': {'shape': 'Payload'}},
            },
            model.ShapeResolver(
                {
                    'Payload': {
                        'type': 'structure',
                        'members': [],
                        'eventstream': True,
                    }
                }
            ),
        )
        with self.assertRaises(parsers.ResponseParserError):
            response_dict = {
                'status_code': 200,
                'headers': {},
                'body': RawResponse(b''),
                'context': {'operation_name': 'TestOperation'},
            }
            parser.parse(response_dict, output_shape)

    def test_metadata_always_exists_for_json(self):
        # ResponseMetadata is used for more than just the request id. It
        # should always get populated, even if the request doesn't seem to
        # have an id.
        parser = parsers.JSONParser()
        response = b'{"Str": "mystring"}'
        headers = {}
        output_shape = model.StructureShape(
            'OutputShape',
            {
                'type': 'structure',
                'members': {
                    'Str': {
                        'shape': 'StringType',
                    }
                },
            },
            model.ShapeResolver({'StringType': {'type': 'string'}}),
        )
        parsed = parser.parse(
            {'body': response, 'headers': headers, 'status_code': 200},
            output_shape,
        )
        expected = {
            'Str': 'mystring',
            'ResponseMetadata': {
                'HTTPStatusCode': 200,
                'HTTPHeaders': headers,
            },
        }
        self.assertEqual(parsed, expected)

    def test_response_metadata_on_rest_json_response(self):
        parser = parsers.RestJSONParser()
        response = b'{"Str": "mystring"}'
        headers = {'x-amzn-requestid': 'request-id'}
        output_shape = model.StructureShape(
            'OutputShape',
            {
                'type': 'structure',
                'members': {
                    'Str': {
                        'shape': 'StringType',
                    }
                },
            },
            model.ShapeResolver({'StringType': {'type': 'string'}}),
        )
        parsed = parser.parse(
            {'body': response, 'headers': headers, 'status_code': 200},
            output_shape,
        )
        # Note that the response metadata is normalized to match the query
        # protocol, even though this is not how it appears in the output.
        self.assertEqual(
            parsed,
            {
                'Str': 'mystring',
                'ResponseMetadata': {
                    'RequestId': 'request-id',
                    'HTTPStatusCode': 200,
                    'HTTPHeaders': headers,
                },
            },
        )

    def test_metadata_always_exists_on_rest_json_response(self):
        # ResponseMetadata is used for more than just the request id. It
        # should always get populated, even if the request doesn't seem to
        # have an id.
        parser = parsers.RestJSONParser()
        response = b'{"Str": "mystring"}'
        headers = {}
        output_shape = model.StructureShape(
            'OutputShape',
            {
                'type': 'structure',
                'members': {
                    'Str': {
                        'shape': 'StringType',
                    }
                },
            },
            model.ShapeResolver({'StringType': {'type': 'string'}}),
        )
        parsed = parser.parse(
            {'body': response, 'headers': headers, 'status_code': 200},
            output_shape,
        )
        expected = {
            'Str': 'mystring',
            'ResponseMetadata': {
                'HTTPStatusCode': 200,
                'HTTPHeaders': headers,
            },
        }
        self.assertEqual(parsed, expected)

    def test_response_metadata_from_s3_response(self):
        # Even though s3 is a rest-xml service, it's response metadata
        # is slightly different.  It has two request ids, both come from
        # the response headers, are both are named differently from other
        # rest-xml responses.
        headers = {'x-amz-id-2': 'second-id', 'x-amz-request-id': 'request-id'}
        parser = parsers.RestXMLParser()
        parsed = parser.parse(
            {'body': '', 'headers': headers, 'status_code': 200}, None
        )
        self.assertEqual(
            parsed,
            {
                'ResponseMetadata': {
                    'RequestId': 'request-id',
                    'HostId': 'second-id',
                    'HTTPStatusCode': 200,
                    'HTTPHeaders': headers,
                }
            },
        )

    def test_metadata_always_exists_on_rest_xml_response(self):
        # ResponseMetadata is used for more than just the request id. It
        # should always get populated, even if the request doesn't seem to
        # have an id.
        headers = {}
        parser = parsers.RestXMLParser()
        parsed = parser.parse(
            {'body': '', 'headers': headers, 'status_code': 200}, None
        )
        expected = {
            'ResponseMetadata': {'HTTPStatusCode': 200, 'HTTPHeaders': headers}
        }
        self.assertEqual(parsed, expected)

    def test_checksum_metadata_parsed_from_response_context(self):
        headers = {}
        response_dict = {
            'status_code': 200,
            'headers': headers,
            'body': b'',
            'context': {'checksum': {'response_algorithm': 'crc32'}},
        }
        parser = parsers.RestXMLParser()
        parsed = parser.parse(response_dict, None)
        expected_algorithm = 'crc32'
        actual_algorithm = parsed['ResponseMetadata']['ChecksumAlgorithm']
        self.assertEqual(actual_algorithm, expected_algorithm)


class TestTaggedUnions(unittest.TestCase):
    def assert_tagged_union_response_with_unknown_member(
        self,
        parser,
        response,
        output_shape,
        expected_parsed_response,
        expected_log,
    ):
        with self.assertLogs() as captured_log:
            parsed = parser.parse(response, output_shape)
            self.assertEqual(parsed, expected_parsed_response)
            self.assertEqual(len(captured_log.records), 1)
            self.assertIn(
                (
                    'Received a tagged union response with member '
                    'unknown to client'
                ),
                captured_log.records[0].getMessage(),
            )

    def test_base_json_parser_handles_unknown_member(self):
        parser = parsers.JSONParser()
        response = b'{"Foo": "mystring"}'
        headers = {'x-amzn-requestid': 'request-id'}
        output_shape = model.StructureShape(
            'OutputShape',
            {
                'type': 'structure',
                'union': True,
                'members': {
                    'Str': {
                        'shape': 'StringType',
                    }
                },
            },
            model.ShapeResolver({'StringType': {'type': 'string'}}),
        )
        response = {'body': response, 'headers': headers, 'status_code': 200}
        # Parsed response omits data from service since it is not
        # modeled in the client
        expected_parsed_response = {
            'SDK_UNKNOWN_MEMBER': {'name': 'Foo'},
            'ResponseMetadata': {
                'RequestId': 'request-id',
                'HTTPStatusCode': 200,
                'HTTPHeaders': {'x-amzn-requestid': 'request-id'},
            },
        }
        expected_log = "Received a response with an unknown member Foo set"
        self.assert_tagged_union_response_with_unknown_member(
            parser,
            response,
            output_shape,
            expected_parsed_response,
            expected_log,
        )

    def test_base_xml_parser_handles_unknown_member(self):
        parser = parsers.QueryParser()
        response = (
            b'<OperationNameResponse>'
            b'  <OperationNameResult><Foo>mystring</Foo></OperationNameResult>'
            b'  <ResponseMetadata>'
            b'    <RequestId>request-id</RequestId>'
            b'  </ResponseMetadata>'
            b'</OperationNameResponse>'
        )
        output_shape = model.StructureShape(
            'OutputShape',
            {
                'type': 'structure',
                'union': True,
                'resultWrapper': 'OperationNameResult',
                'members': {
                    'Str': {
                        'shape': 'StringType',
                    },
                },
            },
            model.ShapeResolver(
                {
                    'StringType': {
                        'type': 'string',
                    },
                }
            ),
        )
        response = {'body': response, 'headers': {}, 'status_code': 200}
        # Parsed response omits data from service since it is not
        # modeled in the client
        expected_parsed_response = {
            'SDK_UNKNOWN_MEMBER': {'name': 'Foo'},
            'ResponseMetadata': {
                'RequestId': 'request-id',
                'HTTPStatusCode': 200,
                'HTTPHeaders': {},
            },
        }
        expected_log = "Received a response with an unknown member Foo set"
        self.assert_tagged_union_response_with_unknown_member(
            parser,
            response,
            output_shape,
            expected_parsed_response,
            expected_log,
        )

    def test_parser_errors_out_when_multiple_members_set(self):
        parser = parsers.JSONParser()
        response = b'{"Foo": "mystring", "Bar": "mystring2"}'
        headers = {'x-amzn-requestid': 'request-id'}
        output_shape = model.StructureShape(
            'OutputShape',
            {
                'type': 'structure',
                'union': True,
                'members': {
                    'Foo': {
                        'shape': 'StringType',
                    },
                    'Bar': {
                        'shape': 'StringType',
                    },
                },
            },
            model.ShapeResolver({'StringType': {'type': 'string'}}),
        )

        response = {'body': response, 'headers': headers, 'status_code': 200}
        with self.assertRaises(parsers.ResponseParserError):
            parser.parse(response, output_shape)

    def test_parser_accepts_type_metadata_with_union(self):
        parser = parsers.JSONParser()
        response = b'{"Foo": "mystring", "__type": "mytype"}'
        headers = {'x-amzn-requestid': 'request-id'}
        output_shape = model.StructureShape(
            'OutputShape',
            {
                'type': 'structure',
                'union': True,
                'members': {
                    'Foo': {
                        'shape': 'StringType',
                    },
                    'Bar': {
                        'shape': 'StringType',
                    },
                },
            },
            model.ShapeResolver({'StringType': {'type': 'string'}}),
        )

        response = {
            'body': response,
            'headers': headers,
            'status_code': 200,
        }
        parsed = parser.parse(response, output_shape)
        self.assertEqual(parsed['Foo'], 'mystring')


class TestHeaderResponseInclusion(unittest.TestCase):
    def create_parser(self):
        return parsers.JSONParser()

    def create_arbitary_output_shape(self):
        output_shape = model.StructureShape(
            'OutputShape',
            {
                'type': 'structure',
                'members': {
                    'Str': {
                        'shape': 'StringType',
                    }
                },
            },
            model.ShapeResolver({'StringType': {'type': 'string'}}),
        )
        return output_shape

    def test_can_add_errors_into_response(self):
        parser = self.create_parser()
        headers = {
            'x-amzn-requestid': 'request-id',
            'Header1': 'foo',
            'Header2': 'bar',
        }
        output_shape = self.create_arbitary_output_shape()
        parsed = parser.parse(
            {'body': b'{}', 'headers': headers, 'status_code': 200},
            output_shape,
        )
        # The mapped header's keys should all be lower cased
        parsed_headers = {
            'x-amzn-requestid': 'request-id',
            'header1': 'foo',
            'header2': 'bar',
        }
        # Response headers should be mapped as HTTPHeaders.
        self.assertEqual(
            parsed['ResponseMetadata']['HTTPHeaders'], parsed_headers
        )

    def test_can_always_json_serialize_headers(self):
        parser = self.create_parser()
        original_headers = {
            'x-amzn-requestid': 'request-id',
            'Header1': 'foo',
        }
        headers = CustomHeaderDict(original_headers)
        output_shape = self.create_arbitary_output_shape()
        parsed = parser.parse(
            {'body': b'{}', 'headers': headers, 'status_code': 200},
            output_shape,
        )
        metadata = parsed['ResponseMetadata']
        # We've had the contract that you can json serialize a
        # response.  So we want to ensure that despite using a CustomHeaderDict
        # we can always JSON dumps the response metadata.
        self.assertEqual(
            json.loads(json.dumps(metadata))['HTTPHeaders']['header1'], 'foo'
        )


class TestResponseParsingDatetimes(unittest.TestCase):
    def test_can_parse_float_timestamps(self):
        # The type "timestamp" can come back as both an integer or as a float.
        # We need to make sure we handle the case where the timestamp comes
        # back as a float.  It might make sense to move this to protocol tests.
        output_shape = model.Shape(
            shape_name='datetime', shape_model={'type': 'timestamp'}
        )
        parser = parsers.JSONParser()
        timestamp_as_float = b'1407538750.49'
        expected_parsed = datetime.datetime(
            2014, 8, 8, 22, 59, 10, 490000, tzinfo=tzutc()
        )
        parsed = parser.parse(
            {'body': timestamp_as_float, 'headers': [], 'status_code': 200},
            output_shape,
        )
        self.assertEqual(parsed, expected_parsed)


class TestResponseParserFactory(unittest.TestCase):
    def setUp(self):
        self.factory = parsers.ResponseParserFactory()

    def test_rest_parser(self):
        parser = self.factory.create_parser('rest-xml')
        self.assertTrue(isinstance(parser, parsers.BaseRestParser))
        self.assertTrue(isinstance(parser, parsers.BaseXMLResponseParser))

    def test_json_parser(self):
        parser = self.factory.create_parser('json')
        self.assertTrue(isinstance(parser, parsers.BaseJSONParser))


class TestCanDecorateResponseParsing(unittest.TestCase):
    def setUp(self):
        self.factory = parsers.ResponseParserFactory()

    def create_request_dict(self, with_body):
        return {'body': with_body, 'headers': [], 'status_code': 200}

    def test_normal_blob_parsing(self):
        output_shape = model.Shape(
            shape_name='BlobType', shape_model={'type': 'blob'}
        )
        parser = self.factory.create_parser('json')

        hello_world_b64 = b'"aGVsbG8gd29ybGQ="'
        expected_parsed = b'hello world'
        parsed = parser.parse(
            self.create_request_dict(with_body=hello_world_b64), output_shape
        )
        self.assertEqual(parsed, expected_parsed)

    def test_can_decorate_scalar_parsing(self):
        output_shape = model.Shape(
            shape_name='BlobType', shape_model={'type': 'blob'}
        )
        # Here we're overriding the blob parser so that
        # we can change it to a noop parser.
        self.factory.set_parser_defaults(blob_parser=lambda x: x)
        parser = self.factory.create_parser('json')

        hello_world_b64 = b'"aGVsbG8gd29ybGQ="'
        expected_parsed = "aGVsbG8gd29ybGQ="
        parsed = parser.parse(
            self.create_request_dict(with_body=hello_world_b64), output_shape
        )
        self.assertEqual(parsed, expected_parsed)

    def test_can_decorate_timestamp_parser(self):
        output_shape = model.Shape(
            shape_name='datetime', shape_model={'type': 'timestamp'}
        )
        # Here we're overriding the timestamp parser so that
        # we can change it to just convert a string to an integer
        # instead of converting to a datetime.
        self.factory.set_parser_defaults(timestamp_parser=lambda x: int(x))
        parser = self.factory.create_parser('json')

        timestamp_as_int = b'1407538750'
        expected_parsed = int(timestamp_as_int)
        parsed = parser.parse(
            self.create_request_dict(with_body=timestamp_as_int), output_shape
        )
        self.assertEqual(parsed, expected_parsed)


class TestHandlesNoOutputShape(unittest.TestCase):
    """Verify that each protocol handles no output shape properly."""

    def test_empty_rest_json_response(self):
        headers = {'x-amzn-requestid': 'request-id'}
        parser = parsers.RestJSONParser()
        output_shape = None
        parsed = parser.parse(
            {'body': b'', 'headers': headers, 'status_code': 200}, output_shape
        )
        self.assertEqual(
            parsed,
            {
                'ResponseMetadata': {
                    'RequestId': 'request-id',
                    'HTTPStatusCode': 200,
                    'HTTPHeaders': headers,
                }
            },
        )

    def test_empty_rest_xml_response(self):
        # This is the format used by cloudfront, route53.
        headers = {'x-amzn-requestid': 'request-id'}
        parser = parsers.RestXMLParser()
        output_shape = None
        parsed = parser.parse(
            {'body': b'', 'headers': headers, 'status_code': 200}, output_shape
        )
        self.assertEqual(
            parsed,
            {
                'ResponseMetadata': {
                    'RequestId': 'request-id',
                    'HTTPStatusCode': 200,
                    'HTTPHeaders': headers,
                }
            },
        )

    def test_empty_query_response(self):
        body = (
            b'<DeleteTagsResponse xmlns="http://autoscaling.amazonaws.com/">'
            b'  <ResponseMetadata>'
            b'    <RequestId>request-id</RequestId>'
            b'  </ResponseMetadata>'
            b'</DeleteTagsResponse>'
        )
        parser = parsers.QueryParser()
        output_shape = None
        parsed = parser.parse(
            {'body': body, 'headers': {}, 'status_code': 200}, output_shape
        )
        self.assertEqual(
            parsed,
            {
                'ResponseMetadata': {
                    'RequestId': 'request-id',
                    'HTTPStatusCode': 200,
                    'HTTPHeaders': {},
                }
            },
        )

    def test_empty_json_response(self):
        headers = {'x-amzn-requestid': 'request-id'}
        # Output shape of None represents no output shape in the model.
        output_shape = None
        parser = parsers.JSONParser()
        parsed = parser.parse(
            {'body': b'', 'headers': headers, 'status_code': 200}, output_shape
        )
        self.assertEqual(
            parsed,
            {
                'ResponseMetadata': {
                    'RequestId': 'request-id',
                    'HTTPStatusCode': 200,
                    'HTTPHeaders': headers,
                }
            },
        )


class TestHandlesInvalidXMLResponses(unittest.TestCase):
    def test_invalid_xml_shown_in_error_message(self):
        # Missing the closing XML tags.
        invalid_xml = (
            b'<DeleteTagsResponse xmlns="http://autoscaling.amazonaws.com/">'
            b'  <ResponseMetadata>'
        )
        parser = parsers.QueryParser()
        output_shape = None
        # The XML body should be in the error message.
        with self.assertRaisesRegex(
            parsers.ResponseParserError, '<DeleteTagsResponse'
        ):
            parser.parse(
                {'body': invalid_xml, 'headers': {}, 'status_code': 200},
                output_shape,
            )


class TestRESTXMLResponses(unittest.TestCase):
    def test_multiple_structures_list_returns_struture(self):
        # This is to handle the scenario when something is modeled
        # as a structure and instead a list of structures is returned.
        # For this case, a single element from the list should be parsed
        # For botocore, this will be the first element.
        # Currently, this logic may happen in s3's GetBucketLifecycle
        # operation.
        headers = {}
        parser = parsers.RestXMLParser()
        body = (
            '<?xml version="1.0" ?>'
            '<OperationName xmlns="http://s3.amazonaws.com/doc/2006-03-01/">'
            '	<Foo><Bar>first_value</Bar></Foo>'
            '	<Foo><Bar>middle_value</Bar></Foo>'
            '	<Foo><Bar>last_value</Bar></Foo>'
            '</OperationName>'
        )
        builder = model.DenormalizedStructureBuilder()
        output_shape = builder.with_members(
            {
                'Foo': {
                    'type': 'structure',
                    'members': {
                        'Bar': {
                            'type': 'string',
                        }
                    },
                }
            }
        ).build_model()
        parsed = parser.parse(
            {'body': body, 'headers': headers, 'status_code': 200},
            output_shape,
        )
        # Ensure the first element is used out of the list.
        self.assertEqual(parsed['Foo'], {'Bar': 'first_value'})


class TestEventStreamParsers(unittest.TestCase):
    def setUp(self):
        self.parser = parsers.EventStreamXMLParser()
        self.output_shape = model.StructureShape(
            'EventStream',
            {
                'eventstream': True,
                'type': 'structure',
                'members': {
                    'EventA': {'shape': 'EventAStructure'},
                    'EventB': {'shape': 'EventBStructure'},
                    'EventC': {'shape': 'EventCStructure'},
                    'EventD': {'shape': 'EventDStructure'},
                    'EventException': {'shape': 'ExceptionShape'},
                },
            },
            model.ShapeResolver(
                {
                    'EventAStructure': {
                        'event': True,
                        'type': 'structure',
                        'members': {
                            'Stats': {
                                'shape': 'StatsStructure',
                                'eventpayload': True,
                            },
                            'Header': {
                                'shape': 'IntShape',
                                'eventheader': True,
                            },
                        },
                    },
                    'EventBStructure': {
                        'event': True,
                        'type': 'structure',
                        'members': {
                            'Body': {
                                'shape': 'BlobShape',
                                'eventpayload': True,
                            }
                        },
                    },
                    'EventCStructure': {
                        'event': True,
                        'type': 'structure',
                        'members': {
                            'Body': {
                                'shape': 'StringShape',
                                'eventpayload': True,
                            }
                        },
                    },
                    'EventDStructure': {
                        'event': True,
                        'type': 'structure',
                        'members': {
                            'StringField': {'shape': 'StringShape'},
                            'IntField': {'shape': 'IntShape'},
                            'Header': {
                                'shape': 'IntShape',
                                'eventheader': True,
                            },
                        },
                    },
                    'StatsStructure': {
                        'type': 'structure',
                        'members': {
                            'StringField': {'shape': 'StringShape'},
                            'IntField': {'shape': 'IntShape'},
                        },
                    },
                    'BlobShape': {'type': 'blob'},
                    'StringShape': {'type': 'string'},
                    'IntShape': {'type': 'integer'},
                    'ExceptionShape': {
                        'exception': True,
                        'type': 'structure',
                        'members': {'message': {'shape': 'StringShape'}},
                    },
                }
            ),
        )

    def parse_event(self, headers=None, body=None, status_code=200):
        response_dict = {
            'body': body,
            'headers': headers,
            'status_code': status_code,
        }
        return self.parser.parse(response_dict, self.output_shape)

    def test_parses_event_xml(self):
        headers = {'Header': 123, ':event-type': 'EventA'}
        body = (
            b'<Stats xmlns="">'
            b'  <StringField>abcde</StringField>'
            b'  <IntField>1234</IntField>'
            b'</Stats>'
        )
        parsed = self.parse_event(headers, body)
        expected = {
            'EventA': {
                'Header': 123,
                'Stats': {'StringField': 'abcde', 'IntField': 1234},
            }
        }
        self.assertEqual(parsed, expected)

    def test_parses_event_bad_xml(self):
        headers = {'Header': 123, ':event-type': 'EventA'}
        parsed = self.parse_event(headers, b'')
        expected = {'EventA': {'Header': 123, 'Stats': {}}}
        self.assertEqual(parsed, expected)

    def test_parses_event_blob(self):
        headers = {':event-type': 'EventB'}
        parsed = self.parse_event(headers, b'blob')
        expected = {'EventB': {'Body': b'blob'}}
        self.assertEqual(parsed, expected)

    def test_parses_event_string(self):
        headers = {':event-type': 'EventC'}
        parsed = self.parse_event(headers, b'blob')
        expected = {'EventC': {'Body': 'blob'}}
        self.assertEqual(parsed, expected)

    def test_parses_payload_implicit(self):
        headers = {'Header': 123, ':event-type': 'EventD'}
        body = (
            b'<EventD xmlns="">'
            b'  <StringField>abcde</StringField>'
            b'  <IntField>1234</IntField>'
            b'</EventD>'
        )
        parsed = self.parse_event(headers, body)
        expected = {
            'EventD': {'Header': 123, 'StringField': 'abcde', 'IntField': 1234}
        }
        self.assertEqual(parsed, expected)

    def test_parses_error_event(self):
        error_code = 'client/SomeError'
        error_message = 'You did something wrong'
        headers = {
            ':message-type': 'error',
            ':error-code': error_code,
            ':error-message': error_message,
        }
        body = b''
        parsed = self.parse_event(headers, body, status_code=400)
        expected = {'Error': {'Code': error_code, 'Message': error_message}}
        self.assertEqual(parsed, expected)

    def test_parses_exception_event(self):
        self.parser = parsers.EventStreamJSONParser()
        error_code = 'EventException'
        headers = {
            ':message-type': 'exception',
            ':exception-type': error_code,
        }
        body = b'{"message": "You did something wrong"}'
        parsed = self.parse_event(headers, body, status_code=400)
        expected = {
            'Error': {'Code': error_code, 'Message': 'You did something wrong'}
        }
        self.assertEqual(parsed, expected)

    def test_parses_event_json(self):
        self.parser = parsers.EventStreamJSONParser()
        headers = {':event-type': 'EventD'}
        body = b'{' b'  "StringField": "abcde",' b'  "IntField": 1234' b'}'
        parsed = self.parse_event(headers, body)
        expected = {'EventD': {'StringField': 'abcde', 'IntField': 1234}}
        self.assertEqual(parsed, expected)


class TestParseErrorResponses(unittest.TestCase):
    # This class consolidates all the error parsing tests
    # across all the protocols.  We may potentially pull
    # this into the shared protocol tests in the future,
    # so consolidating them into a single class will make
    # this easier.
    def setUp(self):
        self.error_shape = model.StructureShape(
            'ErrorShape',
            {
                'type': 'structure',
                'exception': True,
                'members': {
                    'ModeledField': {
                        'shape': 'StringType',
                    }
                },
            },
            model.ShapeResolver({'StringType': {'type': 'string'}}),
        )

    def test_response_metadata_errors_for_json_protocol(self):
        parser = parsers.JSONParser()
        response = {
            "body": b"""
                {"__type":"amazon.foo.validate#ValidationException",
                 "message":"this is a message"}
                """,
            "status_code": 400,
            "headers": {"x-amzn-requestid": "request-id"},
        }
        parsed = parser.parse(response, None)
        # Even (especially) on an error condition, the
        # ResponseMetadata should be populated.
        self.assertIn('ResponseMetadata', parsed)
        self.assertEqual(parsed['ResponseMetadata']['RequestId'], 'request-id')

        self.assertIn('Error', parsed)
        self.assertEqual(parsed['Error']['Message'], 'this is a message')
        self.assertEqual(parsed['Error']['Code'], 'ValidationException')

    def test_response_metadata_errors_alternate_form_json_protocol(self):
        # Sometimes there is no '#' in the __type.  We need to be
        # able to parse this error message as well.
        parser = parsers.JSONParser()
        response = {
            "body": b"""
                {"__type":"ValidationException",
                 "message":"this is a message"}
                """,
            "status_code": 400,
            "headers": {"x-amzn-requestid": "request-id"},
        }
        parsed = parser.parse(response, None)
        self.assertIn('Error', parsed)
        self.assertEqual(parsed['Error']['Message'], 'this is a message')
        self.assertEqual(parsed['Error']['Code'], 'ValidationException')

    def test_response_with_query_error_for_json_protocol(self):
        parser = parsers.JSONParser()
        response = {
            "body": b"""
                {"__type":"amazon.foo.validate#ValidationException",
                 "message":"this is a message"}
                """,
            "status_code": 400,
            "headers": {
                "x-amzn-requestid": "request-id",
                "x-amzn-query-error": "AWS.SimpleQueueService.NonExistentQueue;Sender",
            },
        }
        parsed = parser.parse(response, None)
        # ResponseMetadata should always be populated.
        self.assertIn('ResponseMetadata', parsed)
        self.assertEqual(parsed['ResponseMetadata']['RequestId'], 'request-id')

        self.assertIn('Error', parsed)
        self.assertEqual(parsed['Error']['Message'], 'this is a message')
        self.assertEqual(
            parsed['Error']['Code'], 'AWS.SimpleQueueService.NonExistentQueue'
        )
        self.assertEqual(
            parsed['Error']['QueryErrorCode'], "ValidationException"
        )
        self.assertEqual(parsed['Error']['Type'], 'Sender')

    def test_response_with_invalid_query_error_for_json_protocol(self):
        parser = parsers.JSONParser()
        response = {
            "body": b"""
                {"__type":"amazon.foo.validate#ValidationException",
                 "message":"this is a message"}
                """,
            "status_code": 400,
            "headers": {
                "x-amzn-requestid": "request-id",
                "x-amzn-query-error": "AWS.SimpleQueueService.NonExistentQueue;sender;400",
            },
        }
        parsed = parser.parse(response, None)
        self.assertIn('ResponseMetadata', parsed)
        self.assertEqual(parsed['ResponseMetadata']['RequestId'], 'request-id')

        self.assertIn('Error', parsed)
        self.assertEqual(parsed['Error']['Message'], 'this is a message')
        self.assertEqual(parsed['Error']['Code'], 'ValidationException')
        self.assertNotIn('QueryErrorCode', parsed['Error'])
        self.assertNotIn('Type', parsed['Error'])

    def test_response_with_incomplete_query_error_for_json_protocol(self):
        parser = parsers.JSONParser()
        response = {
            "body": b"""
                {"__type":"amazon.foo.validate#ValidationException",
                 "message":"this is a message"}
                """,
            "status_code": 400,
            "headers": {
                "x-amzn-requestid": "request-id",
                "x-amzn-query-error": ";sender",
            },
        }
        parsed = parser.parse(response, None)
        self.assertIn('ResponseMetadata', parsed)
        self.assertEqual(parsed['ResponseMetadata']['RequestId'], 'request-id')

        self.assertIn('Error', parsed)
        self.assertEqual(parsed['Error']['Message'], 'this is a message')
        self.assertEqual(parsed['Error']['Code'], 'ValidationException')
        self.assertNotIn('QueryErrorCode', parsed['Error'])
        self.assertNotIn('Type', parsed['Error'])

    def test_response_with_empty_query_errors_for_json_protocol(self):
        parser = parsers.JSONParser()
        response = {
            "body": b"""
                {"__type":"amazon.foo.validate#ValidationException",
                 "message":"this is a message"}
                """,
            "status_code": 400,
            "headers": {
                "x-amzn-requestid": "request-id",
                "x-amzn-query-error": "",
            },
        }
        parsed = parser.parse(response, None)
        self.assertIn('ResponseMetadata', parsed)
        self.assertEqual(parsed['ResponseMetadata']['RequestId'], 'request-id')

        self.assertIn('Error', parsed)
        self.assertEqual(parsed['Error']['Message'], 'this is a message')
        self.assertEqual(parsed['Error']['Code'], 'ValidationException')
        self.assertNotIn('QueryErrorCode', parsed['Error'])
        self.assertNotIn('Type', parsed['Error'])

    def test_parse_error_response_for_query_protocol(self):
        body = (
            b'<ErrorResponse xmlns="https://iam.amazonaws.com/doc/2010-05-08/">'
            b'  <Error>'
            b'    <Type>Sender</Type>'
            b'    <Code>InvalidInput</Code>'
            b'    <Message>ARN asdf is not valid.</Message>'
            b'  </Error>'
            b'  <RequestId>request-id</RequestId>'
            b'</ErrorResponse>'
        )
        parser = parsers.QueryParser()
        parsed = parser.parse(
            {'body': body, 'headers': {}, 'status_code': 400}, None
        )
        self.assertIn('Error', parsed)
        self.assertEqual(
            parsed['Error'],
            {
                'Code': 'InvalidInput',
                'Message': 'ARN asdf is not valid.',
                'Type': 'Sender',
            },
        )

    def test_can_parse_sdb_error_response_query_protocol(self):
        body = (
            b'<OperationNameResponse>'
            b'    <Errors>'
            b'        <Error>'
            b'            <Code>1</Code>'
            b'            <Message>msg</Message>'
            b'        </Error>'
            b'    </Errors>'
            b'    <RequestId>abc-123</RequestId>'
            b'</OperationNameResponse>'
        )
        parser = parsers.QueryParser()
        parsed = parser.parse(
            {'body': body, 'headers': {}, 'status_code': 500}, None
        )
        self.assertIn('Error', parsed)
        self.assertEqual(parsed['Error'], {'Code': '1', 'Message': 'msg'})
        self.assertEqual(
            parsed['ResponseMetadata'],
            {'RequestId': 'abc-123', 'HTTPStatusCode': 500, 'HTTPHeaders': {}},
        )

    def test_can_parser_ec2_errors(self):
        body = (
            b'<Response>'
            b'  <Errors>'
            b'    <Error>'
            b'      <Code>InvalidInstanceID.NotFound</Code>'
            b'      <Message>The instance ID i-12345 does not exist</Message>'
            b'    </Error>'
            b'  </Errors>'
            b'  <RequestID>06f382b0-d521-4bb6-988c-ca49d5ae6070</RequestID>'
            b'</Response>'
        )
        parser = parsers.EC2QueryParser()
        parsed = parser.parse(
            {'body': body, 'headers': {}, 'status_code': 400}, None
        )
        self.assertIn('Error', parsed)
        self.assertEqual(
            parsed['Error'],
            {
                'Code': 'InvalidInstanceID.NotFound',
                'Message': 'The instance ID i-12345 does not exist',
            },
        )

    def test_can_parse_rest_xml_errors(self):
        body = (
            b'<ErrorResponse xmlns="https://route53.amazonaws.com/doc/2013-04-01/">'
            b'  <Error>'
            b'    <Type>Sender</Type>'
            b'    <Code>NoSuchHostedZone</Code>'
            b'    <Message>No hosted zone found with ID: foobar</Message>'
            b'  </Error>'
            b'  <RequestId>bc269cf3-d44f-11e5-8779-2d21c30eb3f1</RequestId>'
            b'</ErrorResponse>'
        )
        parser = parsers.RestXMLParser()
        parsed = parser.parse(
            {'body': body, 'headers': {}, 'status_code': 400}, None
        )
        self.assertIn('Error', parsed)
        self.assertEqual(
            parsed['Error'],
            {
                'Code': 'NoSuchHostedZone',
                'Message': 'No hosted zone found with ID: foobar',
                'Type': 'Sender',
            },
        )

    def test_can_parse_rest_json_errors(self):
        body = b'{"Message":"Function not found: foo","Type":"User"}'
        headers = {
            'x-amzn-requestid': 'request-id',
            'x-amzn-errortype': 'ResourceNotFoundException:http://url/',
        }
        parser = parsers.RestJSONParser()
        parsed = parser.parse(
            {'body': body, 'headers': headers, 'status_code': 400}, None
        )
        self.assertIn('Error', parsed)
        self.assertEqual(
            parsed['Error'],
            {
                'Code': 'ResourceNotFoundException',
                'Message': 'Function not found: foo',
            },
        )

    def test_error_response_with_no_body_rest_json(self):
        parser = parsers.RestJSONParser()
        response = b''
        headers = {'content-length': '0', 'connection': 'keep-alive'}
        output_shape = None
        parsed = parser.parse(
            {'body': response, 'headers': headers, 'status_code': 504},
            output_shape,
        )

        self.assertIn('Error', parsed)
        self.assertEqual(
            parsed['Error'], {'Code': '504', 'Message': 'Gateway Timeout'}
        )
        self.assertEqual(
            parsed['ResponseMetadata'],
            {'HTTPStatusCode': 504, 'HTTPHeaders': headers},
        )

    def test_error_response_with_string_body_rest_json(self):
        parser = parsers.RestJSONParser()
        response = b'HTTP content length exceeded 1049600 bytes.'
        headers = {'content-length': '0', 'connection': 'keep-alive'}
        output_shape = None
        parsed = parser.parse(
            {'body': response, 'headers': headers, 'status_code': 413},
            output_shape,
        )

        self.assertIn('Error', parsed)
        self.assertEqual(
            parsed['Error'],
            {'Code': '413', 'Message': response.decode('utf-8')},
        )
        self.assertEqual(
            parsed['ResponseMetadata'],
            {'HTTPStatusCode': 413, 'HTTPHeaders': headers},
        )

    def test_error_response_with_xml_body_rest_json(self):
        parser = parsers.RestJSONParser()
        response = (
            b'<AccessDeniedException>'
            b'   <Message>Unable to determine service/operation name to be authorized</Message>'
            b'</AccessDeniedException>'
        )
        headers = {'content-length': '0', 'connection': 'keep-alive'}
        output_shape = None
        parsed = parser.parse(
            {'body': response, 'headers': headers, 'status_code': 403},
            output_shape,
        )

        self.assertIn('Error', parsed)
        self.assertEqual(
            parsed['Error'],
            {'Code': '403', 'Message': response.decode('utf-8')},
        )
        self.assertEqual(
            parsed['ResponseMetadata'],
            {'HTTPStatusCode': 403, 'HTTPHeaders': headers},
        )

    def test_s3_error_response(self):
        body = (
            b'<Error>'
            b'  <Code>NoSuchBucket</Code>'
            b'  <Message>error message</Message>'
            b'  <BucketName>asdf</BucketName>'
            b'  <RequestId>EF1EF43A74415102</RequestId>'
            b'  <HostId>hostid</HostId>'
            b'</Error>'
        )
        headers = {'x-amz-id-2': 'second-id', 'x-amz-request-id': 'request-id'}
        parser = parsers.RestXMLParser()
        parsed = parser.parse(
            {'body': body, 'headers': headers, 'status_code': 400}, None
        )
        self.assertIn('Error', parsed)
        self.assertEqual(
            parsed['Error'],
            {
                'Code': 'NoSuchBucket',
                'Message': 'error message',
                'BucketName': 'asdf',
                # We don't want the RequestId/HostId because they're already
                # present in the ResponseMetadata key.
            },
        )
        self.assertEqual(
            parsed['ResponseMetadata'],
            {
                'RequestId': 'request-id',
                'HostId': 'second-id',
                'HTTPStatusCode': 400,
                'HTTPHeaders': headers,
            },
        )

    def test_s3_error_response_with_no_body(self):
        # If you try to HeadObject a key that does not exist,
        # you will get an empty body.  When this happens
        # we expect that we will use Code/Message from the
        # HTTP status code.
        body = ''
        headers = {'x-amz-id-2': 'second-id', 'x-amz-request-id': 'request-id'}
        parser = parsers.RestXMLParser()
        parsed = parser.parse(
            {'body': body, 'headers': headers, 'status_code': 404}, None
        )
        self.assertIn('Error', parsed)
        self.assertEqual(
            parsed['Error'],
            {
                'Code': '404',
                'Message': 'Not Found',
            },
        )
        self.assertEqual(
            parsed['ResponseMetadata'],
            {
                'RequestId': 'request-id',
                'HostId': 'second-id',
                'HTTPStatusCode': 404,
                'HTTPHeaders': headers,
            },
        )

    def test_can_parse_glacier_error_response(self):
        body = (
            b'{"code":"AccessDeniedException","type":"Client","message":'
            b'"Access denied"}'
        )
        headers = {'x-amzn-requestid': 'request-id'}
        parser = parsers.RestJSONParser()
        parsed = parser.parse(
            {'body': body, 'headers': headers, 'status_code': 400}, None
        )
        self.assertEqual(
            parsed['Error'],
            {'Message': 'Access denied', 'Code': 'AccessDeniedException'},
        )

    def test_can_parse_restjson_error_code(self):
        body = b'''{
            "status": "error",
            "errors": [{"message": "[*Deprecated*: blah"}],
            "adds": 0,
            "__type": "#WasUnableToParseThis",
            "message": "blah",
            "deletes": 0}'''
        headers = {'x-amzn-requestid': 'request-id'}
        parser = parsers.RestJSONParser()
        parsed = parser.parse(
            {'body': body, 'headers': headers, 'status_code': 400}, None
        )
        self.assertEqual(
            parsed['Error'],
            {'Message': 'blah', 'Code': 'WasUnableToParseThis'},
        )

    def test_can_parse_with_case_insensitive_keys(self):
        body = (
            b'{"Code":"AccessDeniedException","type":"Client","Message":'
            b'"Access denied"}'
        )
        headers = {'x-amzn-requestid': 'request-id'}
        parser = parsers.RestJSONParser()
        parsed = parser.parse(
            {'body': body, 'headers': headers, 'status_code': 400}, None
        )
        self.assertEqual(
            parsed['Error'],
            {'Message': 'Access denied', 'Code': 'AccessDeniedException'},
        )

    def test_can_parse_rest_json_modeled_fields(self):
        body = (
            b'{"ModeledField":"Some modeled field",'
            b'"Message":"Some message"}'
        )
        parser = parsers.RestJSONParser()
        response_dict = {
            'status_code': 400,
            'headers': {},
            'body': body,
        }
        parsed = parser.parse(response_dict, self.error_shape)
        expected_parsed = {
            'ModeledField': 'Some modeled field',
        }
        self.assertEqual(parsed, expected_parsed)

    def test_can_parse_rest_xml_modeled_fields(self):
        parser = parsers.RestXMLParser()
        body = (
            b'<?xml version="1.0"?>\n<ErrorResponse xmlns="http://foo.bar">'
            b'<Error><Type>Sender</Type><Code>NoSuchDistribution</Code>'
            b'<Message>The specified distribution does not exist.</Message>'
            b'<ModeledField>Some modeled field</ModeledField>'
            b'</Error>'
            b'</ErrorResponse>'
        )
        response_dict = {
            'status_code': 400,
            'headers': {},
            'body': body,
        }
        parsed = parser.parse(response_dict, self.error_shape)
        expected_parsed = {
            'ModeledField': 'Some modeled field',
        }
        self.assertEqual(parsed, expected_parsed)

    def test_can_parse_ec2_modeled_fields(self):
        body = (
            b'<Response><Errors><Error>'
            b'<Code>ExceptionShape</Code>'
            b'<Message>Foo message</Message>'
            b'<ModeledField>Some modeled field</ModeledField>'
            b'</Error></Errors></Response>'
        )
        parser = parsers.EC2QueryParser()
        response_dict = {
            'status_code': 400,
            'headers': {},
            'body': body,
        }
        parsed = parser.parse(response_dict, self.error_shape)
        expected_parsed = {
            'ModeledField': 'Some modeled field',
        }
        self.assertEqual(parsed, expected_parsed)

    def test_can_parse_query_modeled_fields(self):
        parser = parsers.QueryParser()
        body = (
            b'<?xml version="1.0"?>\n<ErrorResponse xmlns="http://foo.bar">'
            b'<Error><Type>Sender</Type><Code>SomeCode</Code>'
            b'<Message>A message</Message>'
            b'<ModeledField>Some modeled field</ModeledField>'
            b'</Error>'
            b'</ErrorResponse>'
        )
        response_dict = {
            'status_code': 400,
            'headers': {},
            'body': body,
        }
        parsed = parser.parse(response_dict, self.error_shape)
        expected_parsed = {
            'ModeledField': 'Some modeled field',
        }
        self.assertEqual(parsed, expected_parsed)

    def test_can_parse_json_modeled_fields(self):
        body = (
            b'{"ModeledField":"Some modeled field",'
            b'"Message":"Some message",'
            b'"__type": "Prefix#SomeError"}'
        )
        parser = parsers.JSONParser()
        response_dict = {
            'status_code': 400,
            'headers': {},
            'body': body,
        }
        parsed = parser.parse(response_dict, self.error_shape)
        expected_parsed = {
            'ModeledField': 'Some modeled field',
        }
        self.assertEqual(parsed, expected_parsed)

    def test_can_parse_route53_with_missing_message(self):
        # The message isn't always in the XML response (or even the headers).
        # We should be able to handle this gracefully and still at least
        # populate a "Message" key so that consumers don't have to
        # conditionally check for this.
        body = (
            b'<ErrorResponse>'
            b'  <Error>'
            b'    <Type>Sender</Type>'
            b'    <Code>InvalidInput</Code>'
            b'  </Error>'
            b'  <RequestId>id</RequestId>'
            b'</ErrorResponse>'
        )
        parser = parsers.RestXMLParser()
        parsed = parser.parse(
            {'body': body, 'headers': {}, 'status_code': 400}, None
        )
        error = parsed['Error']
        self.assertEqual(error['Code'], 'InvalidInput')
        # Even though there's no <Message /> we should
        # still populate an empty string.
        self.assertEqual(error['Message'], '')


def _generic_test_bodies():
    generic_html_body = (
        b'<html><body><b>Http/1.1 Service Unavailable</b></body></html>'
    )
    empty_body = b''
    none_body = None

    return [generic_html_body, empty_body, none_body]


@pytest.mark.parametrize(
    "parser, body",
    itertools.product(
        parsers.PROTOCOL_PARSERS.values(), _generic_test_bodies()
    ),
)
def test_can_handle_generic_error_message(parser, body):
    # There are times when you can get a service to respond with a generic
    # html error page.  We should be able to handle this case.
    parsed = parser().parse(
        {'body': body, 'headers': {}, 'status_code': 503}, None
    )
    assert parsed['Error'] == {'Code': '503', 'Message': 'Service Unavailable'}
    assert parsed['ResponseMetadata']['HTTPStatusCode'] == 503



# Copyright 2012-2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import base64
import json
import logging
from itertools import tee

import jmespath

from botocore.exceptions import PaginationError
from botocore.utils import merge_dicts, set_value_from_jmespath

log = logging.getLogger(__name__)


class TokenEncoder:
    """Encodes dictionaries into opaque strings.

    This for the most part json dumps + base64 encoding, but also supports
    having bytes in the dictionary in addition to the types that json can
    handle by default.

    This is intended for use in encoding pagination tokens, which in some
    cases can be complex structures and / or contain bytes.
    """

    def encode(self, token):
        """Encodes a dictionary to an opaque string.

        :type token: dict
        :param token: A dictionary containing pagination information,
            particularly the service pagination token(s) but also other boto
            metadata.

        :rtype: str
        :returns: An opaque string
        """
        try:
            # Try just using json dumps first to avoid having to traverse
            # and encode the dict. In 99.9999% of cases this will work.
            json_string = json.dumps(token)
        except (TypeError, UnicodeDecodeError):
            # If normal dumping failed, go through and base64 encode all bytes.
            encoded_token, encoded_keys = self._encode(token, [])

            # Save the list of all the encoded key paths. We can safely
            # assume that no service will ever use this key.
            encoded_token['boto_encoded_keys'] = encoded_keys

            # Now that the bytes are all encoded, dump the json.
            json_string = json.dumps(encoded_token)

        # base64 encode the json string to produce an opaque token string.
        return base64.b64encode(json_string.encode('utf-8')).decode('utf-8')

    def _encode(self, data, path):
        """Encode bytes in given data, keeping track of the path traversed."""
        if isinstance(data, dict):
            return self._encode_dict(data, path)
        elif isinstance(data, list):
            return self._encode_list(data, path)
        elif isinstance(data, bytes):
            return self._encode_bytes(data, path)
        else:
            return data, []

    def _encode_list(self, data, path):
        """Encode any bytes in a list, noting the index of what is encoded."""
        new_data = []
        encoded = []
        for i, value in enumerate(data):
            new_path = path + [i]
            new_value, new_encoded = self._encode(value, new_path)
            new_data.append(new_value)
            encoded.extend(new_encoded)
        return new_data, encoded

    def _encode_dict(self, data, path):
        """Encode any bytes in a dict, noting the index of what is encoded."""
        new_data = {}
        encoded = []
        for key, value in data.items():
            new_path = path + [key]
            new_value, new_encoded = self._encode(value, new_path)
            new_data[key] = new_value
            encoded.extend(new_encoded)
        return new_data, encoded

    def _encode_bytes(self, data, path):
        """Base64 encode a byte string."""
        return base64.b64encode(data).decode('utf-8'), [path]


class TokenDecoder:
    """Decodes token strings back into dictionaries.

    This performs the inverse operation to the TokenEncoder, accepting
    opaque strings and decoding them into a useable form.
    """

    def decode(self, token):
        """Decodes an opaque string to a dictionary.

        :type token: str
        :param token: A token string given by the botocore pagination
            interface.

        :rtype: dict
        :returns: A dictionary containing pagination information,
            particularly the service pagination token(s) but also other boto
            metadata.
        """
        json_string = base64.b64decode(token.encode('utf-8')).decode('utf-8')
        decoded_token = json.loads(json_string)

        # Remove the encoding metadata as it is read since it will no longer
        # be needed.
        encoded_keys = decoded_token.pop('boto_encoded_keys', None)
        if encoded_keys is None:
            return decoded_token
        else:
            return self._decode(decoded_token, encoded_keys)

    def _decode(self, token, encoded_keys):
        """Find each encoded value and decode it."""
        for key in encoded_keys:
            encoded = self._path_get(token, key)
            decoded = base64.b64decode(encoded.encode('utf-8'))
            self._path_set(token, key, decoded)
        return token

    def _path_get(self, data, path):
        """Return the nested data at the given path.

        For instance:
            data = {'foo': ['bar', 'baz']}
            path = ['foo', 0]
            ==> 'bar'
        """
        # jmespath isn't used here because it would be difficult to actually
        # create the jmespath query when taking all of the unknowns of key
        # structure into account. Gross though this is, it is simple and not
        # very error prone.
        d = data
        for step in path:
            d = d[step]
        return d

    def _path_set(self, data, path, value):
        """Set the value of a key in the given data.

        Example:
            data = {'foo': ['bar', 'baz']}
            path = ['foo', 1]
            value = 'bin'
            ==> data = {'foo': ['bar', 'bin']}
        """
        container = self._path_get(data, path[:-1])
        container[path[-1]] = value


class PaginatorModel:
    def __init__(self, paginator_config):
        self._paginator_config = paginator_config['pagination']

    def get_paginator(self, operation_name):
        try:
            single_paginator_config = self._paginator_config[operation_name]
        except KeyError:
            raise ValueError(
                "Paginator for operation does not exist: %s" % operation_name
            )
        return single_paginator_config


class PageIterator:
    """An iterable object to paginate API results.
    Please note it is NOT a python iterator.
    Use ``iter`` to wrap this as a generator.
    """

    def __init__(
        self,
        method,
        input_token,
        output_token,
        more_results,
        result_keys,
        non_aggregate_keys,
        limit_key,
        max_items,
        starting_token,
        page_size,
        op_kwargs,
    ):
        self._method = method
        self._input_token = input_token
        self._output_token = output_token
        self._more_results = more_results
        self._result_keys = result_keys
        self._max_items = max_items
        self._limit_key = limit_key
        self._starting_token = starting_token
        self._page_size = page_size
        self._op_kwargs = op_kwargs
        self._resume_token = None
        self._non_aggregate_key_exprs = non_aggregate_keys
        self._non_aggregate_part = {}
        self._token_encoder = TokenEncoder()
        self._token_decoder = TokenDecoder()

    @property
    def result_keys(self):
        return self._result_keys

    @property
    def resume_token(self):
        """Token to specify to resume pagination."""
        return self._resume_token

    @resume_token.setter
    def resume_token(self, value):
        if not isinstance(value, dict):
            raise ValueError("Bad starting token: %s" % value)

        if 'boto_truncate_amount' in value:
            token_keys = sorted(self._input_token + ['boto_truncate_amount'])
        else:
            token_keys = sorted(self._input_token)
        dict_keys = sorted(value.keys())

        if token_keys == dict_keys:
            self._resume_token = self._token_encoder.encode(value)
        else:
            raise ValueError("Bad starting token: %s" % value)

    @property
    def non_aggregate_part(self):
        return self._non_aggregate_part

    def __iter__(self):
        current_kwargs = self._op_kwargs
        previous_next_token = None
        next_token = {key: None for key in self._input_token}
        if self._starting_token is not None:
            # If the starting token exists, populate the next_token with the
            # values inside it. This ensures that we have the service's
            # pagination token on hand if we need to truncate after the
            # first response.
            next_token = self._parse_starting_token()[0]
        # The number of items from result_key we've seen so far.
        total_items = 0
        first_request = True
        primary_result_key = self.result_keys[0]
        starting_truncation = 0
        self._inject_starting_params(current_kwargs)
        while True:
            response = self._make_request(current_kwargs)
            parsed = self._extract_parsed_response(response)
            if first_request:
                # The first request is handled differently.  We could
                # possibly have a resume/starting token that tells us where
                # to index into the retrieved page.
                if self._starting_token is not None:
                    starting_truncation = self._handle_first_request(
                        parsed, primary_result_key, starting_truncation
                    )
                first_request = False
                self._record_non_aggregate_key_values(parsed)
            else:
                # If this isn't the first request, we have already sliced into
                # the first request and had to make additional requests after.
                # We no longer need to add this to truncation.
                starting_truncation = 0
            current_response = primary_result_key.search(parsed)
            if current_response is None:
                current_response = []
            num_current_response = len(current_response)
            truncate_amount = 0
            if self._max_items is not None:
                truncate_amount = (
                    total_items + num_current_response - self._max_items
                )
            if truncate_amount > 0:
                self._truncate_response(
                    parsed,
                    primary_result_key,
                    truncate_amount,
                    starting_truncation,
                    next_token,
                )
                yield response
                break
            else:
                yield response
                total_items += num_current_response
                next_token = self._get_next_token(parsed)
                if all(t is None for t in next_token.values()):
                    break
                if (
                    self._max_items is not None
                    and total_items == self._max_items
                ):
                    # We're on a page boundary so we can set the current
                    # next token to be the resume token.
                    self.resume_token = next_token
                    break
                if (
                    previous_next_token is not None
                    and previous_next_token == next_token
                ):
                    message = (
                        f"The same next token was received "
                        f"twice: {next_token}"
                    )
                    raise PaginationError(message=message)
                self._inject_token_into_kwargs(current_kwargs, next_token)
                previous_next_token = next_token

    def search(self, expression):
        """Applies a JMESPath expression to a paginator

        Each page of results is searched using the provided JMESPath
        expression. If the result is not a list, it is yielded
        directly. If the result is a list, each element in the result
        is yielded individually (essentially implementing a flatmap in
        which the JMESPath search is the mapping function).

        :type expression: str
        :param expression: JMESPath expression to apply to each page.

        :return: Returns an iterator that yields the individual
            elements of applying a JMESPath expression to each page of
            results.
        """
        compiled = jmespath.compile(expression)
        for page in self:
            results = compiled.search(page)
            if isinstance(results, list):
                yield from results
            else:
                # Yield result directly if it is not a list.
                yield results

    def _make_request(self, current_kwargs):
        return self._method(**current_kwargs)

    def _extract_parsed_response(self, response):
        return response

    def _record_non_aggregate_key_values(self, response):
        non_aggregate_keys = {}
        for expression in self._non_aggregate_key_exprs:
            result = expression.search(response)
            set_value_from_jmespath(
                non_aggregate_keys, expression.expression, result
            )
        self._non_aggregate_part = non_aggregate_keys

    def _inject_starting_params(self, op_kwargs):
        # If the user has specified a starting token we need to
        # inject that into the operation's kwargs.
        if self._starting_token is not None:
            # Don't need to do anything special if there is no starting
            # token specified.
            next_token = self._parse_starting_token()[0]
            self._inject_token_into_kwargs(op_kwargs, next_token)
        if self._page_size is not None:
            # Pass the page size as the parameter name for limiting
            # page size, also known as the limit_key.
            op_kwargs[self._limit_key] = self._page_size

    def _inject_token_into_kwargs(self, op_kwargs, next_token):
        for name, token in next_token.items():
            if (token is not None) and (token != 'None'):
                op_kwargs[name] = token
            elif name in op_kwargs:
                del op_kwargs[name]

    def _handle_first_request(
        self, parsed, primary_result_key, starting_truncation
    ):
        # If the payload is an array or string, we need to slice into it
        # and only return the truncated amount.
        starting_truncation = self._parse_starting_token()[1]
        all_data = primary_result_key.search(parsed)
        if isinstance(all_data, (list, str)):
            data = all_data[starting_truncation:]
        else:
            data = None
        set_value_from_jmespath(parsed, primary_result_key.expression, data)
        # We also need to truncate any secondary result keys
        # because they were not truncated in the previous last
        # response.
        for token in self.result_keys:
            if token == primary_result_key:
                continue
            sample = token.search(parsed)
            if isinstance(sample, list):
                empty_value = []
            elif isinstance(sample, str):
                empty_value = ''
            elif isinstance(sample, (int, float)):
                empty_value = 0
            else:
                empty_value = None
            set_value_from_jmespath(parsed, token.expression, empty_value)
        return starting_truncation

    def _truncate_response(
        self,
        parsed,
        primary_result_key,
        truncate_amount,
        starting_truncation,
        next_token,
    ):
        original = primary_result_key.search(parsed)
        if original is None:
            original = []
        amount_to_keep = len(original) - truncate_amount
        truncated = original[:amount_to_keep]
        set_value_from_jmespath(
            parsed, primary_result_key.expression, truncated
        )
        # The issue here is that even though we know how much we've truncated
        # we need to account for this globally including any starting
        # left truncation. For example:
        # Raw response: [0,1,2,3]
        # Starting index: 1
        # Max items: 1
        # Starting left truncation: [1, 2, 3]
        # End right truncation for max items: [1]
        # However, even though we only kept 1, this is post
        # left truncation so the next starting index should be 2, not 1
        # (left_truncation + amount_to_keep).
        next_token['boto_truncate_amount'] = (
            amount_to_keep + starting_truncation
        )
        self.resume_token = next_token

    def _get_next_token(self, parsed):
        if self._more_results is not None:
            if not self._more_results.search(parsed):
                return {}
        next_tokens = {}
        for output_token, input_key in zip(
            self._output_token, self._input_token
        ):
            next_token = output_token.search(parsed)
            # We do not want to include any empty strings as actual tokens.
            # Treat them as None.
            if next_token:
                next_tokens[input_key] = next_token
            else:
                next_tokens[input_key] = None
        return next_tokens

    def result_key_iters(self):
        teed_results = tee(self, len(self.result_keys))
        return [
            ResultKeyIterator(i, result_key)
            for i, result_key in zip(teed_results, self.result_keys)
        ]

    def build_full_result(self):
        complete_result = {}
        for response in self:
            page = response
            # We want to try to catch operation object pagination
            # and format correctly for those. They come in the form
            # of a tuple of two elements: (http_response, parsed_responsed).
            # We want the parsed_response as that is what the page iterator
            # uses. We can remove it though once operation objects are removed.
            if isinstance(response, tuple) and len(response) == 2:
                page = response[1]
            # We're incrementally building the full response page
            # by page.  For each page in the response we need to
            # inject the necessary components from the page
            # into the complete_result.
            for result_expression in self.result_keys:
                # In order to incrementally update a result key
                # we need to search the existing value from complete_result,
                # then we need to search the _current_ page for the
                # current result key value.  Then we append the current
                # value onto the existing value, and re-set that value
                # as the new value.
                result_value = result_expression.search(page)
                if result_value is None:
                    continue
                existing_value = result_expression.search(complete_result)
                if existing_value is None:
                    # Set the initial result
                    set_value_from_jmespath(
                        complete_result,
                        result_expression.expression,
                        result_value,
                    )
                    continue
                # Now both result_value and existing_value contain something
                if isinstance(result_value, list):
                    existing_value.extend(result_value)
                elif isinstance(result_value, (int, float, str)):
                    # Modify the existing result with the sum or concatenation
                    set_value_from_jmespath(
                        complete_result,
                        result_expression.expression,
                        existing_value + result_value,
                    )
        merge_dicts(complete_result, self.non_aggregate_part)
        if self.resume_token is not None:
            complete_result['NextToken'] = self.resume_token
        return complete_result

    def _parse_starting_token(self):
        if self._starting_token is None:
            return None

        # The starting token is a dict passed as a base64 encoded string.
        next_token = self._starting_token
        try:
            next_token = self._token_decoder.decode(next_token)
            index = 0
            if 'boto_truncate_amount' in next_token:
                index = next_token.get('boto_truncate_amount')
                del next_token['boto_truncate_amount']
        except (ValueError, TypeError):
            next_token, index = self._parse_starting_token_deprecated()
        return next_token, index

    def _parse_starting_token_deprecated(self):
        """
        This handles parsing of old style starting tokens, and attempts to
        coerce them into the new style.
        """
        log.debug(
            "Attempting to fall back to old starting token parser. For "
            "token: %s" % self._starting_token
        )
        if self._starting_token is None:
            return None

        parts = self._starting_token.split('___')
        next_token = []
        index = 0
        if len(parts) == len(self._input_token) + 1:
            try:
                index = int(parts.pop())
            except ValueError:
                # This doesn't look like a valid old-style token, so we're
                # passing it along as an opaque service token.
                parts = [self._starting_token]

        for part in parts:
            if part == 'None':
                next_token.append(None)
            else:
                next_token.append(part)
        return self._convert_deprecated_starting_token(next_token), index

    def _convert_deprecated_starting_token(self, deprecated_token):
        """
        This attempts to convert a deprecated starting token into the new
        style.
        """
        len_deprecated_token = len(deprecated_token)
        len_input_token = len(self._input_token)
        if len_deprecated_token > len_input_token:
            raise ValueError("Bad starting token: %s" % self._starting_token)
        elif len_deprecated_token < len_input_token:
            log.debug(
                "Old format starting token does not contain all input "
                "tokens. Setting the rest, in order, as None."
            )
            for i in range(len_input_token - len_deprecated_token):
                deprecated_token.append(None)
        return dict(zip(self._input_token, deprecated_token))


class Paginator:
    PAGE_ITERATOR_CLS = PageIterator

    def __init__(self, method, pagination_config, model):
        self._model = model
        self._method = method
        self._pagination_cfg = pagination_config
        self._output_token = self._get_output_tokens(self._pagination_cfg)
        self._input_token = self._get_input_tokens(self._pagination_cfg)
        self._more_results = self._get_more_results_token(self._pagination_cfg)
        self._non_aggregate_keys = self._get_non_aggregate_keys(
            self._pagination_cfg
        )
        self._result_keys = self._get_result_keys(self._pagination_cfg)
        self._limit_key = self._get_limit_key(self._pagination_cfg)

    @property
    def result_keys(self):
        return self._result_keys

    def _get_non_aggregate_keys(self, config):
        keys = []
        for key in config.get('non_aggregate_keys', []):
            keys.append(jmespath.compile(key))
        return keys

    def _get_output_tokens(self, config):
        output = []
        output_token = config['output_token']
        if not isinstance(output_token, list):
            output_token = [output_token]
        for config in output_token:
            output.append(jmespath.compile(config))
        return output

    def _get_input_tokens(self, config):
        input_token = self._pagination_cfg['input_token']
        if not isinstance(input_token, list):
            input_token = [input_token]
        return input_token

    def _get_more_results_token(self, config):
        more_results = config.get('more_results')
        if more_results is not None:
            return jmespath.compile(more_results)

    def _get_result_keys(self, config):
        result_key = config.get('result_key')
        if result_key is not None:
            if not isinstance(result_key, list):
                result_key = [result_key]
            result_key = [jmespath.compile(rk) for rk in result_key]
            return result_key

    def _get_limit_key(self, config):
        return config.get('limit_key')

    def paginate(self, **kwargs):
        """Create paginator object for an operation.

        This returns an iterable object.  Iterating over
        this object will yield a single page of a response
        at a time.

        """
        page_params = self._extract_paging_params(kwargs)
        return self.PAGE_ITERATOR_CLS(
            self._method,
            self._input_token,
            self._output_token,
            self._more_results,
            self._result_keys,
            self._non_aggregate_keys,
            self._limit_key,
            page_params['MaxItems'],
            page_params['StartingToken'],
            page_params['PageSize'],
            kwargs,
        )

    def _extract_paging_params(self, kwargs):
        pagination_config = kwargs.pop('PaginationConfig', {})
        max_items = pagination_config.get('MaxItems', None)
        if max_items is not None:
            max_items = int(max_items)
        page_size = pagination_config.get('PageSize', None)
        if page_size is not None:
            if self._limit_key is None:
                raise PaginationError(
                    message="PageSize parameter is not supported for the "
                    "pagination interface for this operation."
                )
            input_members = self._model.input_shape.members
            limit_key_shape = input_members.get(self._limit_key)
            if limit_key_shape.type_name == 'string':
                if not isinstance(page_size, str):
                    page_size = str(page_size)
            else:
                page_size = int(page_size)
        return {
            'MaxItems': max_items,
            'StartingToken': pagination_config.get('StartingToken', None),
            'PageSize': page_size,
        }


class ResultKeyIterator:
    """Iterates over the results of paginated responses.

    Each iterator is associated with a single result key.
    Iterating over this object will give you each element in
    the result key list.

    :param pages_iterator: An iterator that will give you
        pages of results (a ``PageIterator`` class).
    :param result_key: The JMESPath expression representing
        the result key.

    """

    def __init__(self, pages_iterator, result_key):
        self._pages_iterator = pages_iterator
        self.result_key = result_key

    def __iter__(self):
        for page in self._pages_iterator:
            results = self.result_key.search(page)
            if results is None:
                results = []
            yield from results



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from datetime import datetime
from math import ceil

import pytest

from botocore.paginate import TokenDecoder, TokenEncoder
from botocore.stub import StubAssertionError, Stubber
from tests import BaseSessionTest, random_chars


class TestRDSPagination(BaseSessionTest):
    def setUp(self):
        super().setUp()
        self.region = 'us-west-2'
        self.client = self.session.create_client('rds', self.region)
        self.stubber = Stubber(self.client)

    def test_can_specify_zero_marker(self):
        service_response = {
            'LogFileData': 'foo',
            'Marker': '2',
            'AdditionalDataPending': True,
        }
        expected_params = {
            'DBInstanceIdentifier': 'foo',
            'LogFileName': 'bar',
            'NumberOfLines': 2,
            'Marker': '0',
        }
        function_name = 'download_db_log_file_portion'

        # The stubber will assert that the function is called with the expected
        # parameters.
        self.stubber.add_response(
            function_name, service_response, expected_params
        )
        self.stubber.activate()

        try:
            paginator = self.client.get_paginator(function_name)
            result = paginator.paginate(
                DBInstanceIdentifier='foo',
                LogFileName='bar',
                NumberOfLines=2,
                PaginationConfig={'StartingToken': '0', 'MaxItems': 3},
            ).build_full_result()
            self.assertEqual(result['LogFileData'], 'foo')
            self.assertIn('NextToken', result)
        except StubAssertionError as e:
            self.fail(str(e))


class TestAutoscalingPagination(BaseSessionTest):
    def setUp(self):
        super().setUp()
        self.region = 'us-west-2'
        self.client = self.session.create_client(
            'autoscaling',
            self.region,
            aws_secret_access_key='foo',
            aws_access_key_id='bar',
            aws_session_token='baz',
        )
        self.stubber = Stubber(self.client)
        self.stubber.activate()

    def _setup_scaling_pagination(
        self, page_size=200, max_items=100, total_items=600
    ):
        """
        Add to the stubber to test paginating describe_scaling_activities.

        WARNING: This only handles cases where max_items cleanly divides
        page_size.
        """
        requests_per_page = page_size / max_items
        if requests_per_page != ceil(requests_per_page):
            raise NotImplementedError(
                "This only handles setup where max_items is less than "
                "page_size and where max_items evenly divides page_size."
            )
        requests_per_page = int(requests_per_page)
        num_pages = int(ceil(total_items / page_size))

        previous_next_token = None
        for i in range(num_pages):
            page = self.create_describe_scaling_response(page_size=page_size)

            # Don't create a next_token for the final page
            if i + 1 == num_pages:
                next_token = None
            else:
                next_token = random_chars(10)

            expected_args = {}
            if previous_next_token:
                expected_args['StartingToken'] = previous_next_token

            # The same page may be accessed multiple times because we are
            # truncating it at max_items
            for _ in range(requests_per_page - 1):
                # The page is copied because the paginator will modify the
                # response object, causing issues when using the stubber.
                self.stubber.add_response(
                    'describe_scaling_activities', page.copy()
                )

            if next_token is not None:
                page['NextToken'] = next_token

            # Copying the page here isn't necessary because it is about to
            # be blown away anyway.
            self.stubber.add_response('describe_scaling_activities', page)

            previous_next_token = next_token

    def create_describe_scaling_response(self, page_size=200):
        """Create a valid describe_scaling_activities response."""
        page = []
        date = datetime.now()
        for _ in range(page_size):
            page.append(
                {
                    'AutoScalingGroupName': 'test',
                    'ActivityId': random_chars(10),
                    'Cause': 'test',
                    'StartTime': date,
                    'StatusCode': '200',
                }
            )
        return {'Activities': page}

    def test_repeated_build_full_results(self):
        # This ensures that we can cleanly paginate using build_full_results.
        max_items = 100
        total_items = 600
        self._setup_scaling_pagination(
            max_items=max_items, total_items=total_items, page_size=200
        )
        paginator = self.client.get_paginator('describe_scaling_activities')
        conf = {'MaxItems': max_items}

        pagination_tokens = []

        result = paginator.paginate(PaginationConfig=conf).build_full_result()
        all_results = result['Activities']
        while 'NextToken' in result:
            starting_token = result['NextToken']
            # We should never get a duplicate pagination token.
            self.assertNotIn(starting_token, pagination_tokens)
            pagination_tokens.append(starting_token)

            conf['StartingToken'] = starting_token
            pages = paginator.paginate(PaginationConfig=conf)
            result = pages.build_full_result()
            all_results.extend(result['Activities'])

        self.assertEqual(len(all_results), total_items)


class TestCloudwatchLogsPagination(BaseSessionTest):
    def setUp(self):
        super().setUp()
        self.region = 'us-west-2'
        self.client = self.session.create_client(
            'logs',
            self.region,
            aws_secret_access_key='foo',
            aws_access_key_id='bar',
            aws_session_token='baz',
        )
        self.stubber = Stubber(self.client)
        self.stubber.activate()

    def test_token_with_triple_underscores(self):
        response = {
            'events': [
                {
                    'logStreamName': 'foobar',
                    'timestamp': 1560195817,
                    'message': 'a thing happened',
                    'ingestionTime': 1560195817,
                    'eventId': 'foo',
                }
            ],
            'searchedLogStreams': [
                {
                    'logStreamName': 'foobar',
                    'searchedCompletely': False,
                }
            ],
        }
        group_name = 'foo'
        token = 'foo___bar'
        expected_args = {
            'logGroupName': group_name,
            'nextToken': token,
        }
        self.stubber.add_response('filter_log_events', response, expected_args)
        paginator = self.client.get_paginator('filter_log_events')
        pages = paginator.paginate(
            PaginationConfig={
                'MaxItems': 1,
                'StartingToken': token,
            },
            logGroupName=group_name,
        )
        result = pages.build_full_result()
        self.assertEqual(len(result['events']), 1)


@pytest.mark.parametrize(
    "token_dict",
    (
        {'foo': 'bar'},
        {'foo': b'bar'},
        {'foo': {'bar': b'baz'}},
        {'foo': ['bar', b'baz']},
        {'foo': b'\xff'},
        {'foo': {'bar': b'baz', 'bin': [b'bam']}},
    ),
)
def test_token_encoding(token_dict):
    encoded = TokenEncoder().encode(token_dict)
    assert isinstance(encoded, str)
    decoded = TokenDecoder().decode(encoded)
    assert decoded == token_dict



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import pytest

from botocore import regions
from botocore.exceptions import EndpointVariantError, NoRegionError
from tests import unittest


class TestEndpointResolver(unittest.TestCase):
    def _template(self):
        return {
            'partitions': [
                {
                    'partition': 'aws',
                    'dnsSuffix': 'amazonaws.com',
                    'regionRegex': r'^(us|eu)\-\w+$',
                    'defaults': {
                        'hostname': '{service}.{region}.{dnsSuffix}',
                        'variants': [
                            {
                                'hostname': '{service}-fips.{region}.{dnsSuffix}',
                                'tags': ['fips'],
                            },
                            {
                                'dnsSuffix': 'api.aws',
                                'hostname': '{service}.{region}.{dnsSuffix}',
                                'tags': ['dualstack'],
                            },
                            {
                                'dnsSuffix': 'api.aws',
                                'hostname': '{service}-fips.{region}.{dnsSuffix}',
                                'tags': ['dualstack', 'fips'],
                            },
                        ],
                    },
                    'regions': {
                        'us-foo': {'regionName': 'a'},
                        'us-bar': {'regionName': 'b'},
                        'eu-baz': {'regionName': 'd'},
                    },
                    'services': {
                        'ec2': {
                            'defaults': {
                                'protocols': ['http', 'https'],
                                'variants': [
                                    {
                                        'dnsSuffix': 'api.aws',
                                        'hostname': 'api.ec2.{region}.{dnsSuffix}',
                                        'tags': ['dualstack'],
                                    }
                                ],
                            },
                            'endpoints': {
                                'us-foo': {
                                    'hostname': 'ec2.us-foo.amazonaws.com',
                                    'variants': [
                                        {
                                            'dnsSuffix': 'api.aws',
                                            'hostname': 'ec2.foo.{dnsSuffix}',
                                            'tags': ['dualstack'],
                                        },
                                        {
                                            'hostname': 'ec2-fips.foo.amazonaws.com',
                                            'tags': ['fips'],
                                        },
                                        {
                                            'hostname': 'ec2-fips.foo.api.aws',
                                            'tags': ['fips', 'dualstack'],
                                        },
                                    ],
                                },
                                'us-bar': {},
                                'us-dep': {
                                    'deprecated': True,
                                },
                                'us-fizz': {
                                    'credentialScope': {'region': 'us-fizz'},
                                    'hostname': 'ec2.us-fizz.amazonaws.com',
                                    'variants': [
                                        {
                                            'hostname': 'ec2.fizz.api.aws',
                                            'tags': ['dualstack'],
                                        }
                                    ],
                                },
                                'eu-baz': {},
                                'd': {},
                            },
                        },
                        's3': {
                            'defaults': {
                                'sslCommonName': '{service}.{region}.{dnsSuffix}',
                                'variants': [
                                    {
                                        'hostname': 's3.dualstack.{region}.{dnsSuffix}',
                                        'tags': ['dualstack'],
                                    },
                                    {
                                        'hostname': 's3-fips.{region}.{dnsSuffix}',
                                        'tags': ['fips'],
                                    },
                                    {
                                        'hostname': 's3-fips.dualstack.{region}.{dnsSuffix}',
                                        'tags': ['dualstack', 'fips'],
                                    },
                                ],
                            },
                            'endpoints': {
                                'us-foo': {
                                    'sslCommonName': '{region}.{service}.{dnsSuffix}'
                                },
                                'us-bar': {},
                                'us-fizz': {
                                    'hostname': 's3.api.us-fizz.amazonaws.com',
                                    'variants': [
                                        {'tags': ['dualstack']},
                                        {'tags': ['fips']},
                                    ],
                                },
                                'eu-baz': {'hostname': 'foo'},
                            },
                        },
                        'not-regionalized': {
                            'isRegionalized': False,
                            'partitionEndpoint': 'aws',
                            'endpoints': {
                                'aws': {'hostname': 'not-regionalized'},
                                'us-foo': {},
                                'eu-baz': {},
                            },
                        },
                        'non-partition': {
                            'partitionEndpoint': 'aws',
                            'endpoints': {
                                'aws': {'hostname': 'host'},
                                'us-foo': {},
                            },
                        },
                        'merge': {
                            'defaults': {
                                'signatureVersions': ['v2'],
                                'protocols': ['http'],
                            },
                            'endpoints': {
                                'us-foo': {'signatureVersions': ['v4']},
                                'us-bar': {'protocols': ['https']},
                            },
                        },
                    },
                },
                {
                    'partition': 'foo',
                    'dnsSuffix': 'foo.com',
                    'regionRegex': r'^(foo)\-\w+$',
                    'defaults': {
                        'hostname': '{service}.{region}.{dnsSuffix}',
                        'protocols': ['http'],
                        'foo': 'bar',
                    },
                    'regions': {
                        'foo-1': {'regionName': '1'},
                        'foo-2': {'regionName': '2'},
                        'foo-3': {'regionName': '3'},
                    },
                    'services': {
                        'ec2': {
                            'endpoints': {
                                'foo-1': {'foo': 'baz'},
                                'foo-2': {},
                                'foo-3': {},
                            }
                        }
                    },
                },
                {
                    'partition': 'aws-iso',
                    'dnsSuffix': 'amazonaws.com',
                    'defaults': {
                        'hostname': '{service}.{region}.{dnsSuffix}',
                        'protocols': ['http'],
                    },
                    'regions': {
                        'foo-1': {'regionName': '1'},
                        'foo-2': {'regionName': '2'},
                        'foo-3': {'regionName': '3'},
                    },
                    'services': {
                        'ec2': {
                            'endpoints': {
                                'foo-1': {'foo': 'baz'},
                                'foo-2': {},
                                'foo-3': {},
                            }
                        }
                    },
                },
            ]
        }

    def test_ensures_region_is_not_none(self):
        with self.assertRaises(NoRegionError):
            resolver = regions.EndpointResolver(self._template())
            resolver.construct_endpoint('foo', None)

    def test_ensures_required_keys_present(self):
        with self.assertRaises(ValueError):
            regions.EndpointResolver({})

    def test_returns_empty_list_when_listing_for_different_partition(self):
        resolver = regions.EndpointResolver(self._template())
        self.assertEqual([], resolver.get_available_endpoints('ec2', 'bar'))

    def test_returns_empty_list_when_no_service_found(self):
        resolver = regions.EndpointResolver(self._template())
        self.assertEqual([], resolver.get_available_endpoints('what?'))

    def test_gets_endpoint_names(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.get_available_endpoints(
            'ec2', allow_non_regional=True
        )
        self.assertEqual(
            ['d', 'eu-baz', 'us-bar', 'us-dep', 'us-fizz', 'us-foo'],
            sorted(result),
        )

    def test_gets_endpoint_names_for_partition(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.get_available_endpoints(
            'ec2', allow_non_regional=True, partition_name='foo'
        )
        self.assertEqual(['foo-1', 'foo-2', 'foo-3'], sorted(result))

    def test_list_regional_endpoints_only(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.get_available_endpoints(
            'ec2', allow_non_regional=False
        )
        self.assertEqual(['eu-baz', 'us-bar', 'us-foo'], sorted(result))

    def test_returns_none_when_no_match(self):
        resolver = regions.EndpointResolver(self._template())
        self.assertIsNone(resolver.construct_endpoint('foo', 'baz'))

    def test_constructs_regionalized_endpoints_for_exact_matches(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('not-regionalized', 'eu-baz')
        self.assertEqual(
            'not-regionalized.eu-baz.amazonaws.com', result['hostname']
        )
        self.assertEqual('aws', result['partition'])
        self.assertEqual('eu-baz', result['endpointName'])

    def test_constructs_partition_endpoints_for_real_partition_region(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('not-regionalized', 'us-bar')
        self.assertEqual('not-regionalized', result['hostname'])
        self.assertEqual('aws', result['partition'])
        self.assertEqual('aws', result['endpointName'])

    def test_constructs_partition_endpoints_for_regex_match(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('not-regionalized', 'us-abc')
        self.assertEqual('not-regionalized', result['hostname'])

    def test_constructs_endpoints_for_regionalized_regex_match(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('s3', 'us-abc')
        self.assertEqual('s3.us-abc.amazonaws.com', result['hostname'])

    def test_constructs_endpoints_for_unknown_service_but_known_region(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('unknown', 'us-foo')
        self.assertEqual('unknown.us-foo.amazonaws.com', result['hostname'])

    def test_merges_service_keys(self):
        resolver = regions.EndpointResolver(self._template())
        us_foo = resolver.construct_endpoint('merge', 'us-foo')
        us_bar = resolver.construct_endpoint('merge', 'us-bar')
        self.assertEqual(['http'], us_foo['protocols'])
        self.assertEqual(['v4'], us_foo['signatureVersions'])
        self.assertEqual(['https'], us_bar['protocols'])
        self.assertEqual(['v2'], us_bar['signatureVersions'])

    def test_merges_partition_default_keys_with_no_overwrite(self):
        resolver = regions.EndpointResolver(self._template())
        resolved = resolver.construct_endpoint('ec2', 'foo-1')
        self.assertEqual('baz', resolved['foo'])
        self.assertEqual(['http'], resolved['protocols'])

    def test_merges_partition_default_keys_with_overwrite(self):
        resolver = regions.EndpointResolver(self._template())
        resolved = resolver.construct_endpoint('ec2', 'foo-2')
        self.assertEqual('bar', resolved['foo'])
        self.assertEqual(['http'], resolved['protocols'])

    def test_gives_hostname_and_common_name_unaltered(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('s3', 'eu-baz')
        self.assertEqual('s3.eu-baz.amazonaws.com', result['sslCommonName'])
        self.assertEqual('foo', result['hostname'])

    def tests_uses_partition_endpoint_when_no_region_provided(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('not-regionalized')
        self.assertEqual('not-regionalized', result['hostname'])
        self.assertEqual('aws', result['endpointName'])

    def test_returns_dns_suffix_if_available(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('not-regionalized')
        self.assertEqual(result['dnsSuffix'], 'amazonaws.com')

    def test_construct_dualstack_from_endpoint_variant(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint(
            'ec2', 'us-foo', use_dualstack_endpoint=True
        )
        self.assertEqual(result['hostname'], 'ec2.foo.api.aws')
        self.assertEqual(result['dnsSuffix'], 'api.aws')

    def test_construct_dualstack_endpoint_from_service_default_variant(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint(
            'ec2', 'us-bar', use_dualstack_endpoint=True
        )
        self.assertEqual(result['hostname'], 'api.ec2.us-bar.api.aws')
        self.assertEqual(result['dnsSuffix'], 'api.aws')

    def test_construct_dualstack_endpoint_from_partition_default_variant(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint(
            'dynamodb', 'us-bar', use_dualstack_endpoint=True
        )
        self.assertEqual(result['hostname'], 'dynamodb.us-bar.api.aws')
        self.assertEqual(result['dnsSuffix'], 'api.aws')

    def test_constructs_dualstack_endpoint_no_hostname_in_variant(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint(
            's3', 'us-fizz', use_dualstack_endpoint=True
        )
        self.assertEqual('s3.dualstack.us-fizz.api.aws', result['hostname'])

    def test_constructs_endpoint_dualstack_no_variant_dns_suffix(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint(
            's3', 'us-bar', use_dualstack_endpoint=True
        )
        self.assertEqual('s3.dualstack.us-bar.api.aws', result['hostname'])

    def test_construct_dualstack_endpoint_iso_partition_raise_exception(self):
        with self.assertRaises(EndpointVariantError):
            resolver = regions.EndpointResolver(self._template())
            resolver.construct_endpoint(
                'foo', 'foo-1', 'aws-iso', use_dualstack_endpoint=True
            )

    def test_get_partition_dns_suffix_no_tags(self):
        resolver = regions.EndpointResolver(self._template())
        self.assertEqual(
            resolver.get_partition_dns_suffix('aws'), 'amazonaws.com'
        )

    def test_get_partition_dualstack_dns_suffix(self):
        resolver = regions.EndpointResolver(self._template())
        self.assertEqual(
            resolver.get_partition_dns_suffix('aws', ['dualstack']), 'api.aws'
        )

    def test_get_partition_dualstack_dns_suffix_does_not_exist(self):
        resolver = regions.EndpointResolver(self._template())
        self.assertIsNone(
            resolver.get_partition_dns_suffix('foo', ['dualstack'])
        )

    def test_get_available_fips_endpoints(self):
        resolver = regions.EndpointResolver(self._template())
        fips_endpoints = resolver.get_available_endpoints(
            'ec2', endpoint_variant_tags=['fips']
        )
        self.assertEqual(fips_endpoints, ['us-foo'])

    def test_get_available_dualstack_endpoints(self):
        resolver = regions.EndpointResolver(self._template())
        dualstack_endpoints = resolver.get_available_endpoints(
            'ec2', endpoint_variant_tags=['dualstack']
        )
        self.assertEqual(dualstack_endpoints, ['us-foo'])

    def test_get_available_fips_and_dualstack_endpoints(self):
        resolver = regions.EndpointResolver(self._template())
        fips_and_dualstack_endpoints = resolver.get_available_endpoints(
            'ec2', endpoint_variant_tags=['fips', 'dualstack']
        )
        self.assertEqual(fips_and_dualstack_endpoints, ['us-foo'])

    def test_get_available_fips_endpoints_none(self):
        resolver = regions.EndpointResolver(self._template())
        fips_endpoints = resolver.get_available_endpoints(
            'ec2', 'foo', endpoint_variant_tags=['fips']
        )
        self.assertEqual(fips_endpoints, [])

    def test_get_available_dualstack_endpoints_none(self):
        resolver = regions.EndpointResolver(self._template())
        dualstack_endpoints = resolver.get_available_endpoints(
            'ec2', 'foo', endpoint_variant_tags=['dualstack']
        )
        self.assertEqual(dualstack_endpoints, [])

    def test_get_available_fips_and_dualstack_endpoints_none(self):
        resolver = regions.EndpointResolver(self._template())
        fips_and_dualstack_endpoints = resolver.get_available_endpoints(
            'ec2', 'foo', endpoint_variant_tags=['fips', 'dualstack']
        )
        self.assertEqual(fips_and_dualstack_endpoints, [])

    def test_construct_deprecated_endpoint_raises_warning(self):
        resolver = regions.EndpointResolver(self._template())
        with self.assertLogs('botocore.regions', level='WARNING') as log:
            result = resolver.construct_endpoint(
                'ec2',
                'us-dep',
                use_fips_endpoint=True,
            )
            self.assertIn('deprecated endpoint', log.output[0])
            self.assertEqual(
                result['hostname'], 'ec2-fips.us-dep.amazonaws.com'
            )
            self.assertEqual(result['dnsSuffix'], 'amazonaws.com')

    def test_construct_fips_from_endpoint_variant(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint(
            'ec2', 'us-foo', use_fips_endpoint=True
        )
        self.assertEqual(result['hostname'], 'ec2-fips.foo.amazonaws.com')
        self.assertEqual(result['dnsSuffix'], 'amazonaws.com')

    def test_construct_fips_endpoint_from_service_default_variant(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint(
            'ec2', 'us-bar', use_fips_endpoint=True
        )
        self.assertEqual(result['hostname'], 'ec2-fips.us-bar.amazonaws.com')
        self.assertEqual(result['dnsSuffix'], 'amazonaws.com')

    def test_construct_fips_endpoint_from_partition_default_variant(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint(
            'dynamodb', 'us-bar', use_fips_endpoint=True
        )
        self.assertEqual(
            result['hostname'], 'dynamodb-fips.us-bar.amazonaws.com'
        )
        self.assertEqual(result['dnsSuffix'], 'amazonaws.com')

    def test_constructs_fips_endpoint_no_hostname_in_variant(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint(
            's3', 'us-fizz', use_fips_endpoint=True
        )
        self.assertEqual('s3-fips.us-fizz.amazonaws.com', result['hostname'])

    def test_construct_dualstack_and_fips_from_endpoint_variant(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint(
            'ec2',
            'us-foo',
            use_dualstack_endpoint=True,
            use_fips_endpoint=True,
        )
        self.assertEqual(result['hostname'], 'ec2-fips.foo.api.aws')
        self.assertEqual(result['dnsSuffix'], 'api.aws')

    def test_construct_dualstack_and_fips_endpoint_from_service_default_variant(
        self,
    ):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint(
            'ec2',
            'us-bar',
            use_dualstack_endpoint=True,
            use_fips_endpoint=True,
        )
        self.assertEqual(result['hostname'], 'ec2-fips.us-bar.api.aws')
        self.assertEqual(result['dnsSuffix'], 'api.aws')

    def test_construct_dualstack_and_fips_endpoint_from_partition_default_variant(
        self,
    ):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint(
            'dynamodb',
            'us-bar',
            use_dualstack_endpoint=True,
            use_fips_endpoint=True,
        )
        self.assertEqual(result['hostname'], 'dynamodb-fips.us-bar.api.aws')
        self.assertEqual(result['dnsSuffix'], 'api.aws')

    def test_constructs_dualstack_and_fips_endpoint_no_hostname_in_variant(
        self,
    ):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint(
            's3',
            'us-fizz',
            use_dualstack_endpoint=True,
            use_fips_endpoint=True,
        )
        self.assertEqual(
            's3-fips.dualstack.us-fizz.api.aws', result['hostname']
        )

    def test_construct_fips_endpoint_no_variant_raise_exception(self):
        with self.assertRaises(EndpointVariantError):
            resolver = regions.EndpointResolver(self._template())
            resolver.construct_endpoint(
                'ec2', 'foo-1', 'foo', use_fips_endpoint=True
            )

    def test_construct_dualstack_endpoint_no_variant_raise_exception(self):
        with self.assertRaises(EndpointVariantError):
            resolver = regions.EndpointResolver(self._template())
            resolver.construct_endpoint(
                'ec2', 'foo-1', 'foo', use_dualstack_endpoint=True
            )

    def test_construct_dualstack_and_fips_endpoint_no_variant_raise_exception(
        self,
    ):
        with self.assertRaises(EndpointVariantError):
            resolver = regions.EndpointResolver(self._template())
            resolver.construct_endpoint(
                'ec2',
                'foo-1',
                'foo',
                use_dualstack_endpoint=True,
                use_fips_endpoint=True,
            )


def _variant_test_definitions():
    return [
        {
            "service": "default-pattern-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": False,
            "endpoint": "default-pattern-service.us-west-2.amazonaws.com",
        },
        {
            "service": "default-pattern-service",
            "region": "us-west-2",
            "fips": True,
            "dualstack": False,
            "endpoint": "default-pattern-service-fips.us-west-2.amazonaws.com",
        },
        {
            "service": "default-pattern-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": False,
            "endpoint": "default-pattern-service.af-south-1.amazonaws.com",
        },
        {
            "service": "default-pattern-service",
            "region": "af-south-1",
            "fips": True,
            "dualstack": False,
            "endpoint": "default-pattern-service-fips.af-south-1.amazonaws.com",
        },
        {
            "service": "global-service",
            "region": "aws-global",
            "fips": False,
            "dualstack": False,
            "endpoint": "global-service.amazonaws.com",
        },
        {
            "service": "global-service",
            "region": "aws-global",
            "fips": True,
            "dualstack": False,
            "endpoint": "global-service-fips.amazonaws.com",
        },
        {
            "service": "global-service",
            "region": "foo",
            "fips": False,
            "dualstack": False,
            "endpoint": "global-service.amazonaws.com",
        },
        {
            "service": "global-service",
            "region": "foo",
            "fips": True,
            "dualstack": False,
            "endpoint": "global-service-fips.amazonaws.com",
        },
        {
            "service": "override-variant-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-service.us-west-2.amazonaws.com",
        },
        {
            "service": "override-variant-service",
            "region": "us-west-2",
            "fips": True,
            "dualstack": False,
            "endpoint": "fips.override-variant-service.us-west-2.new.dns.suffix",
        },
        {
            "service": "override-variant-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-service.af-south-1.amazonaws.com",
        },
        {
            "service": "override-variant-service",
            "region": "af-south-1",
            "fips": True,
            "dualstack": False,
            "endpoint": "fips.override-variant-service.af-south-1.new.dns.suffix",
        },
        {
            "service": "override-variant-dns-suffix-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-dns-suffix-service.us-west-2.amazonaws.com",
        },
        {
            "service": "override-variant-dns-suffix-service",
            "region": "us-west-2",
            "fips": True,
            "dualstack": False,
            "endpoint": "override-variant-dns-suffix-service-fips.us-west-2.new.dns.suffix",
        },
        {
            "service": "override-variant-dns-suffix-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-dns-suffix-service.af-south-1.amazonaws.com",
        },
        {
            "service": "override-variant-dns-suffix-service",
            "region": "af-south-1",
            "fips": True,
            "dualstack": False,
            "endpoint": "override-variant-dns-suffix-service-fips.af-south-1.new.dns.suffix",
        },
        {
            "service": "override-variant-hostname-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-hostname-service.us-west-2.amazonaws.com",
        },
        {
            "service": "override-variant-hostname-service",
            "region": "us-west-2",
            "fips": True,
            "dualstack": False,
            "endpoint": "fips.override-variant-hostname-service.us-west-2.amazonaws.com",
        },
        {
            "service": "override-variant-hostname-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-hostname-service.af-south-1.amazonaws.com",
        },
        {
            "service": "override-variant-hostname-service",
            "region": "af-south-1",
            "fips": True,
            "dualstack": False,
            "endpoint": "fips.override-variant-hostname-service.af-south-1.amazonaws.com",
        },
        {
            "service": "override-endpoint-variant-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-endpoint-variant-service.us-west-2.amazonaws.com",
        },
        {
            "service": "override-endpoint-variant-service",
            "region": "us-west-2",
            "fips": True,
            "dualstack": False,
            "endpoint": "fips.override-endpoint-variant-service.us-west-2.amazonaws.com",
        },
        {
            "service": "override-endpoint-variant-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-endpoint-variant-service.af-south-1.amazonaws.com",
        },
        {
            "service": "override-endpoint-variant-service",
            "region": "af-south-1",
            "fips": True,
            "dualstack": False,
            "endpoint": "override-endpoint-variant-service-fips.af-south-1.amazonaws.com",
        },
        {
            "service": "default-pattern-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": False,
            "endpoint": "default-pattern-service.us-west-2.amazonaws.com",
        },
        {
            "service": "default-pattern-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": True,
            "endpoint": "default-pattern-service.us-west-2.api.aws",
        },
        {
            "service": "default-pattern-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": False,
            "endpoint": "default-pattern-service.af-south-1.amazonaws.com",
        },
        {
            "service": "default-pattern-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": True,
            "endpoint": "default-pattern-service.af-south-1.api.aws",
        },
        {
            "service": "global-service",
            "region": "aws-global",
            "fips": False,
            "dualstack": False,
            "endpoint": "global-service.amazonaws.com",
        },
        {
            "service": "global-service",
            "region": "aws-global",
            "fips": False,
            "dualstack": True,
            "endpoint": "global-service.api.aws",
        },
        {
            "service": "global-service",
            "region": "foo",
            "fips": False,
            "dualstack": False,
            "endpoint": "global-service.amazonaws.com",
        },
        {
            "service": "global-service",
            "region": "foo",
            "fips": False,
            "dualstack": True,
            "endpoint": "global-service.api.aws",
        },
        {
            "service": "override-variant-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-service.us-west-2.amazonaws.com",
        },
        {
            "service": "override-variant-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": True,
            "endpoint": "override-variant-service.dualstack.us-west-2.new.dns.suffix",
        },
        {
            "service": "override-variant-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-service.af-south-1.amazonaws.com",
        },
        {
            "service": "override-variant-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": True,
            "endpoint": "override-variant-service.dualstack.af-south-1.new.dns.suffix",
        },
        {
            "service": "override-variant-dns-suffix-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-dns-suffix-service.us-west-2.amazonaws.com",
        },
        {
            "service": "override-variant-dns-suffix-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": True,
            "endpoint": "override-variant-dns-suffix-service.us-west-2.new.dns.suffix",
        },
        {
            "service": "override-variant-dns-suffix-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-dns-suffix-service.af-south-1.amazonaws.com",
        },
        {
            "service": "override-variant-dns-suffix-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": True,
            "endpoint": "override-variant-dns-suffix-service.af-south-1.new.dns.suffix",
        },
        {
            "service": "override-variant-hostname-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-hostname-service.us-west-2.amazonaws.com",
        },
        {
            "service": "override-variant-hostname-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": True,
            "endpoint": "override-variant-hostname-service.dualstack.us-west-2.api.aws",
        },
        {
            "service": "override-variant-hostname-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-hostname-service.af-south-1.amazonaws.com",
        },
        {
            "service": "override-variant-hostname-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": True,
            "endpoint": "override-variant-hostname-service.dualstack.af-south-1.api.aws",
        },
        {
            "service": "override-endpoint-variant-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-endpoint-variant-service.us-west-2.amazonaws.com",
        },
        {
            "service": "override-endpoint-variant-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": True,
            "endpoint": "override-endpoint-variant-service.dualstack.us-west-2.amazonaws.com",
        },
        {
            "service": "override-endpoint-variant-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-endpoint-variant-service.af-south-1.amazonaws.com",
        },
        {
            "service": "override-endpoint-variant-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": True,
            "endpoint": "override-endpoint-variant-service.af-south-1.api.aws",
        },
        {
            "service": "multi-variant-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": False,
            "endpoint": "multi-variant-service.us-west-2.amazonaws.com",
        },
        {
            "service": "multi-variant-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": True,
            "endpoint": "multi-variant-service.dualstack.us-west-2.api.aws",
        },
        {
            "service": "multi-variant-service",
            "region": "us-west-2",
            "fips": True,
            "dualstack": False,
            "endpoint": "fips.multi-variant-service.us-west-2.amazonaws.com",
        },
        {
            "service": "multi-variant-service",
            "region": "us-west-2",
            "fips": True,
            "dualstack": True,
            "endpoint": "fips.multi-variant-service.dualstack.us-west-2.new.dns.suffix",
        },
        {
            "service": "multi-variant-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": False,
            "endpoint": "multi-variant-service.af-south-1.amazonaws.com",
        },
        {
            "service": "multi-variant-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": True,
            "endpoint": "multi-variant-service.dualstack.af-south-1.api.aws",
        },
        {
            "service": "multi-variant-service",
            "region": "af-south-1",
            "fips": True,
            "dualstack": False,
            "endpoint": "fips.multi-variant-service.af-south-1.amazonaws.com",
        },
        {
            "service": "multi-variant-service",
            "region": "af-south-1",
            "fips": True,
            "dualstack": True,
            "endpoint": "fips.multi-variant-service.dualstack.af-south-1.new.dns.suffix",
        },
    ]


def _modeled_variants_template():
    return {
        "partitions": [
            {
                "defaults": {
                    "hostname": "{service}.{region}.{dnsSuffix}",
                    "protocols": ["https"],
                    "signatureVersions": ["v4"],
                    "variants": [
                        {
                            "dnsSuffix": "amazonaws.com",
                            "hostname": "{service}-fips.{region}.{dnsSuffix}",
                            "tags": ["fips"],
                        },
                        {
                            "dnsSuffix": "api.aws",
                            "hostname": "{service}.{region}.{dnsSuffix}",
                            "tags": ["dualstack"],
                        },
                        {
                            "dnsSuffix": "api.aws",
                            "hostname": "{service}-fips.{region}.{dnsSuffix}",
                            "tags": ["dualstack", "fips"],
                        },
                    ],
                },
                "dnsSuffix": "amazonaws.com",
                "partition": "aws",
                "regionRegex": "^(us|eu|ap|sa|ca|me|af)\\-\\w+\\-\\d+$",
                "regions": {
                    "af-south-1": {"description": "Africa (Cape Town)"},
                    "us-west-2": {"description": "US West (Oregon)"},
                },
                "services": {
                    "default-pattern-service": {
                        "endpoints": {
                            "af-south-1": {},
                            "us-west-2": {
                                "variants": [
                                    {
                                        "hostname": "default-pattern-service-fips.us-west-2.amazonaws.com",
                                        "tags": ["fips"],
                                    },
                                    {
                                        "hostname": "default-pattern-service.us-west-2.api.aws",
                                        "tags": ["dualstack"],
                                    },
                                ]
                            },
                        }
                    },
                    "global-service": {
                        "endpoints": {
                            "aws-global": {
                                "credentialScope": {"region": "us-east-1"},
                                "hostname": "global-service.amazonaws.com",
                                "variants": [
                                    {
                                        "hostname": "global-service-fips.amazonaws.com",
                                        "tags": ["fips"],
                                    },
                                    {
                                        "hostname": "global-service.api.aws",
                                        "tags": ["dualstack"],
                                    },
                                ],
                            }
                        },
                        "isRegionalized": False,
                        "partitionEndpoint": "aws-global",
                    },
                    "override-variant-service": {
                        "defaults": {
                            "variants": [
                                {
                                    "hostname": "fips.{service}.{region}.{dnsSuffix}",
                                    "dnsSuffix": "new.dns.suffix",
                                    "tags": ["fips"],
                                },
                                {
                                    "hostname": "{service}.dualstack.{region}.{dnsSuffix}",
                                    "dnsSuffix": "new.dns.suffix",
                                    "tags": ["dualstack"],
                                },
                            ]
                        },
                        "endpoints": {
                            "af-south-1": {},
                            "us-west-2": {
                                "variants": [
                                    {
                                        "hostname": "fips.override-variant-service.us-west-2.new.dns.suffix",
                                        "tags": ["fips"],
                                    },
                                    {
                                        "hostname": "override-variant-service.dualstack.us-west-2.new.dns.suffix",
                                        "tags": ["dualstack"],
                                    },
                                ]
                            },
                        },
                    },
                    "override-variant-dns-suffix-service": {
                        "defaults": {
                            "variants": [
                                {
                                    "dnsSuffix": "new.dns.suffix",
                                    "tags": ["fips"],
                                },
                                {
                                    "dnsSuffix": "new.dns.suffix",
                                    "tags": ["dualstack"],
                                },
                            ]
                        },
                        "endpoints": {
                            "af-south-1": {},
                            "us-west-2": {
                                "variants": [
                                    {
                                        "hostname": "override-variant-dns-suffix-service-fips.us-west-2.new.dns.suffix",
                                        "tags": ["fips"],
                                    },
                                    {
                                        "hostname": "override-variant-dns-suffix-service.us-west-2.new.dns.suffix",
                                        "tags": ["dualstack"],
                                    },
                                ]
                            },
                        },
                    },
                    "override-variant-hostname-service": {
                        "defaults": {
                            "variants": [
                                {
                                    "hostname": "fips.{service}.{region}.{dnsSuffix}",
                                    "tags": ["fips"],
                                },
                                {
                                    "hostname": "{service}.dualstack.{region}.{dnsSuffix}",
                                    "tags": ["dualstack"],
                                },
                            ]
                        },
                        "endpoints": {
                            "af-south-1": {},
                            "us-west-2": {
                                "variants": [
                                    {
                                        "hostname": "fips.override-variant-hostname-service.us-west-2.amazonaws.com",
                                        "tags": ["fips"],
                                    },
                                    {
                                        "hostname": "override-variant-hostname-service.dualstack.us-west-2.api.aws",
                                        "tags": ["dualstack"],
                                    },
                                ]
                            },
                        },
                    },
                    "override-endpoint-variant-service": {
                        "endpoints": {
                            "af-south-1": {},
                            "us-west-2": {
                                "variants": [
                                    {
                                        "hostname": "fips.override-endpoint-variant-service.us-west-2.amazonaws.com",
                                        "tags": ["fips"],
                                    },
                                    {
                                        "hostname": "override-endpoint-variant-service.dualstack.us-west-2.amazonaws.com",
                                        "tags": ["dualstack"],
                                    },
                                ]
                            },
                        }
                    },
                    "multi-variant-service": {
                        "defaults": {
                            "variants": [
                                {
                                    "hostname": "fips.{service}.{region}.{dnsSuffix}",
                                    "tags": ["fips"],
                                },
                                {
                                    "hostname": "{service}.dualstack.{region}.{dnsSuffix}",
                                    "tags": ["dualstack"],
                                },
                                {
                                    "dnsSuffix": "new.dns.suffix",
                                    "hostname": "fips.{service}.dualstack.{region}.{dnsSuffix}",
                                    "tags": ["fips", "dualstack"],
                                },
                            ]
                        },
                        "endpoints": {
                            "af-south-1": {"deprecated": True},
                            "us-west-2": {
                                "variants": [
                                    {
                                        "hostname": "fips.multi-variant-service.dualstack.us-west-2.new.dns.suffix",
                                        "tags": ["fips", "dualstack"],
                                    }
                                ]
                            },
                        },
                    },
                },
            },
            {
                "defaults": {
                    "hostname": "{service}.{region}.{dnsSuffix}",
                    "protocols": ["https"],
                    "signatureVersions": ["v4"],
                },
                "dnsSuffix": "c2s.ic.gov",
                "partition": "aws-iso",
                "regionRegex": "^us\\-iso\\-\\w+\\-\\d+$",
                "regions": {"us-iso-east-1": {"description": "US ISO East"}},
                "services": {
                    "some-service": {"endpoints": {"us-iso-east-1": {}}}
                },
            },
        ],
        "version": 3,
    }


@pytest.mark.parametrize("test_case", _variant_test_definitions())
def test_modeled_variants(test_case):
    _verify_expected_endpoint(**test_case)


def _verify_expected_endpoint(service, region, fips, dualstack, endpoint):
    resolver = regions.EndpointResolver(_modeled_variants_template())
    resolved = resolver.construct_endpoint(
        service,
        region,
        use_dualstack_endpoint=dualstack,
        use_fips_endpoint=fips,
    )
    # If we can't resolve the region, we attempt to get a
    # global endpoint.
    if not resolved:
        resolved = resolver.construct_endpoint(
            service,
            region,
            partition_name='aws',
            use_dualstack_endpoint=dualstack,
            use_fips_endpoint=fips,
        )
    assert resolved['hostname'] == endpoint


def test_additional_endpoint_data_exists_with_variants():
    resolver = regions.EndpointResolver(_modeled_variants_template())
    resolved = resolver.construct_endpoint(
        'global-service',
        'aws-global',
        use_fips_endpoint=True,
    )
    assert resolved['credentialScope'] == {'region': 'us-east-1'}



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import pytest

from botocore.session import get_session
from botocore.utils import CLIENT_NAME_TO_HYPHENIZED_SERVICE_ID_OVERRIDES

ENDPOINT_PREFIX_OVERRIDE = {
    # entry in endpoints.json -> actual endpoint prefix.
    # The autoscaling-* services actually send requests to the
    # autoscaling service, but they're exposed as separate clients
    # in botocore.
    'autoscaling-plans': 'autoscaling',
    'application-autoscaling': 'autoscaling',
    # For neptune, we send requests to the RDS endpoint.
    'neptune': 'rds',
    'docdb': 'rds',
    # iotevents data endpoints.json and service-2.json don't line up.
    'ioteventsdata': 'data.iotevents',
    'iotsecuredtunneling': 'api.tunneling.iot',
    'iotwireless': 'api.iotwireless',
    'data.iot': 'data-ats.iot',
}

NOT_SUPPORTED_IN_SDK = [
    'mobileanalytics',
    'transcribestreaming',
]


SESSION = get_session()
LOADER = SESSION.get_component('data_loader')
AVAILABLE_SERVICES = LOADER.list_available_services('service-2')


def _known_endpoint_prefixes():
    # The entries in endpoints.json are keyed off of the endpoint
    # prefix.  We don't directly have that data, so we have to load
    # every service model and look up its endpoint prefix in its
    # ``metadata`` section.
    return {
        SESSION.get_service_model(service_name).endpoint_prefix
        for service_name in AVAILABLE_SERVICES
    }


def _computed_endpoint_prefixes():
    # This verifies client names match up with data from the endpoints.json
    # file.  We want to verify that every entry in the endpoints.json
    # file corresponds to a client we can construct via
    # session.create_client(...).
    # So first we get a list of all the service names in the endpoints
    # file.
    endpoints = LOADER.load_data('endpoints')
    # A service can be in multiple partitions so we're using
    # a set here to remove dupes.
    services_in_endpoints_file = set()
    for partition in endpoints['partitions']:
        for service in partition['services']:
            # There are some services we don't support in the SDK
            # so we don't need to add them to the list of services
            # we need to check.
            if service not in NOT_SUPPORTED_IN_SDK:
                services_in_endpoints_file.add(service)

    # Now we go through every known endpoint prefix in the endpoints.json
    # file and ensure it maps to an endpoint prefix we've seen
    # in a service model.
    endpoint_prefixes = []
    for endpoint_prefix in services_in_endpoints_file:
        # Check for an override where we know that an entry
        # in the endpoints.json actually maps to a different endpoint
        # prefix.
        endpoint_prefix = ENDPOINT_PREFIX_OVERRIDE.get(
            endpoint_prefix, endpoint_prefix
        )
        endpoint_prefixes.append(endpoint_prefix)
    return sorted(endpoint_prefixes)


KNOWN_ENDPOINT_PREFIXES = _known_endpoint_prefixes()
COMPUTED_ENDPOINT_PREFIXES = _computed_endpoint_prefixes()


@pytest.mark.parametrize("endpoint_prefix", COMPUTED_ENDPOINT_PREFIXES)
def test_endpoint_matches_service(endpoint_prefix):
    # We need to cross check all computed endpoints against our
    # known values in endpoints.json, to ensure everything lines
    # up correctly.
    assert endpoint_prefix in KNOWN_ENDPOINT_PREFIXES


@pytest.mark.parametrize("service_name", AVAILABLE_SERVICES)
def test_client_name_matches_hyphenized_service_id(service_name):
    """Generates tests for each service to verify that the computed service
    named based on the service id matches the service name used to
    create a client (i.e the directory name in botocore/data)
    unless there is an explicit exception.
    """
    service_model = SESSION.get_service_model(service_name)
    computed_name = service_model.service_id.replace(' ', '-').lower()

    # Handle known exceptions where we have renamed the service directory
    # for one reason or another.
    actual_service_name = CLIENT_NAME_TO_HYPHENIZED_SERVICE_ID_OVERRIDES.get(
        service_name, service_name
    )

    err_msg = (
        f"Actual service name `{actual_service_name}` does not match "
        f"expected service name we computed: `{computed_name}`"
    )
    assert computed_name == actual_service_name, err_msg
